# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permisions and
# limitations under the License.
"""A library of static analysis functions for ASTs."""

from collections.abc import Callable, Iterable
from typing import Optional, Union

from federated_language.compiler import building_block_analysis
from federated_language.compiler import building_blocks
from federated_language.compiler import intrinsic_defs
from federated_language.compiler import transformation_utils
from federated_language.types import computation_types
from federated_language.types import placements
from federated_language.types import type_analysis


_TypeOrTupleOfTypes = Union[
    type[building_blocks.ComputationBuildingBlock],
    tuple[type[building_blocks.ComputationBuildingBlock], ...],
]


def visit_preorder(
    tree: building_blocks.ComputationBuildingBlock,
    function: Callable[[building_blocks.ComputationBuildingBlock], None],
):

  def _visit(building_block):
    function(building_block)
    return building_block, False

  transformation_utils.transform_preorder(tree, _visit)


def visit_postorder(
    tree: building_blocks.ComputationBuildingBlock,
    function: Callable[[building_blocks.ComputationBuildingBlock], None],
):

  def _visit(building_block):
    function(building_block)
    return building_block, False

  transformation_utils.transform_postorder(tree, _visit)


_BuildingBlockPredicate = Callable[
    [building_blocks.ComputationBuildingBlock], bool
]


def count(
    tree: building_blocks.ComputationBuildingBlock,
    predicate: Optional[_BuildingBlockPredicate] = None,
) -> int:
  """Returns the number of building blocks in `tree` matching `predicate`.

  Args:
    tree: A tree of `building_blocks.ComputationBuildingBlock`s to count.
    predicate: An optional Python function that takes a tree as a parameter and
      returns a boolean value. If `None`, all computations are counted.
  """
  counter = 0

  def _fn(building_block):
    nonlocal counter
    if predicate is None or predicate(building_block):
      counter += 1

  visit_postorder(tree, _fn)
  return counter


def contains(
    tree: building_blocks.ComputationBuildingBlock,
    predicate: _BuildingBlockPredicate,
) -> bool:
  """Returns whether or not a building block in `tree` matches `predicate`."""
  return count(tree, predicate) != 0


def contains_only(
    tree: building_blocks.ComputationBuildingBlock,
    predicate: _BuildingBlockPredicate,
) -> bool:
  """Returns whether or not a building block in `tree` matches `predicate`."""
  return not contains(tree, lambda x: not predicate(x))


def check_has_single_placement(
    comp: building_blocks.ComputationBuildingBlock,
    single_placement: placements.PlacementLiteral,
) -> None:
  """Checks that the AST of `comp` contains only `single_placement`.

  Args:
    comp: Instance of `building_blocks.ComputationBuildingBlock`.
    single_placement: Instance of `placements.PlacementLiteral` which should be
      the only placement present under `comp`.

  Raises:
    ValueError: If the AST under `comp` contains any
    `computation_types.FederatedType` other than `single_placement`.
  """

  def _check_single_placement(comp):
    """Checks that the placement in `type_spec` matches `single_placement`."""
    if (
        isinstance(comp.type_signature, computation_types.FederatedType)
        and comp.type_signature.placement is not single_placement
    ):
      raise ValueError(
          'Comp contains a placement other than {}; '
          'placement {} on comp {} inside the structure. '.format(
              single_placement,
              comp.type_signature.placement,
              comp.compact_representation(),
          )
      )

  visit_postorder(comp, _check_single_placement)


def check_contains_only_reducible_intrinsics(
    comp: building_blocks.ComputationBuildingBlock,
):
  """Checks that `comp` contains intrinsics reducible to aggregate or broadcast.

  Args:
    comp: Instance of `building_blocks.ComputationBuildingBlock` to check for
      presence of intrinsics not currently immediately reducible to
      `FEDERATED_AGGREGATE` or `FEDERATED_BROADCAST`, or local processing.

  Raises:
    ValueError: If we encounter an intrinsic under `comp` that is not reducible.
  """
  reducible_uris = (
      intrinsic_defs.FEDERATED_AGGREGATE.uri,
      intrinsic_defs.FEDERATED_APPLY.uri,
      intrinsic_defs.FEDERATED_BROADCAST.uri,
      intrinsic_defs.FEDERATED_EVAL_AT_CLIENTS.uri,
      intrinsic_defs.FEDERATED_EVAL_AT_SERVER.uri,
      intrinsic_defs.FEDERATED_MAP.uri,
      intrinsic_defs.FEDERATED_MAP_ALL_EQUAL.uri,
      intrinsic_defs.FEDERATED_SECURE_SUM_BITWIDTH.uri,
      intrinsic_defs.FEDERATED_SECURE_SUM.uri,
      intrinsic_defs.FEDERATED_VALUE_AT_CLIENTS.uri,
      intrinsic_defs.FEDERATED_VALUE_AT_SERVER.uri,
      intrinsic_defs.FEDERATED_ZIP_AT_CLIENTS.uri,
      intrinsic_defs.FEDERATED_ZIP_AT_SERVER.uri,
  )

  def _check(comp):
    if (
        isinstance(comp, building_blocks.Intrinsic)
        and comp.uri not in reducible_uris
    ):
      raise ValueError(
          'Encountered an Intrinsic not currently reducible to aggregate or '
          'broadcast, the intrinsic {}'.format(comp.compact_representation())
      )

  visit_postorder(comp, _check)


class NonuniqueNameError(ValueError):

  def __init__(self, comp, name):
    self.comp = comp
    self.name = name
    message = (
        f'The name `{name}` is bound multiple times in the computation:\n'
        f'{comp.compact_representation()}'
    )
    super().__init__(message)


def check_has_unique_names(
    comp: building_blocks.ComputationBuildingBlock,
) -> None:
  """Checks that each variable of `comp` is bound at most once.

  Additionally, checks that `comp` does not mask any names which are unbound
  at the top level.

  Args:
    comp: Instance of `building_blocks.ComputationBuildingBlock`.

  Raises:
    NonuniqueNameError: If we encounter a name that is bound multiple times or a
      binding which would shadow an unbound reference.
  """
  # Initializing `names` to unbound names in `comp` ensures that `comp` does not
  # mask any names from its parent scope.
  names = transformation_utils.get_map_of_unbound_references(comp)[comp]

  def _visit_name(name):
    if name in names:
      raise NonuniqueNameError(comp, name)
    names.add(name)

  def _visit(comp):
    if isinstance(comp, building_blocks.Block):
      for name, _ in comp.locals:
        _visit_name(name)
    elif (
        isinstance(comp, building_blocks.Lambda)
        and comp.parameter_type is not None
    ):
      _visit_name(comp.parameter_name)

  visit_postorder(comp, _visit)


def _extract_nodes_consuming(
    tree: building_blocks.ComputationBuildingBlock,
    predicate: _BuildingBlockPredicate,
) -> set[building_blocks.ComputationBuildingBlock]:
  """Returns the set of AST nodes which consume nodes matching `predicate`.

  Notice we adopt the convention that a node which itself satisfies the
  predicate is in this set.

  Args:
    tree: Instance of `building_blocks.ComputationBuildingBlock` to view as an
      abstract syntax tree, and construct the set of nodes in this tree having a
      dependency on nodes matching `predicate`; that is, the set of nodes whose
      value depends on evaluating nodes matching `predicate`.
    predicate: One-arg callable, accepting arguments of type
      `building_blocks.ComputationBuildingBlock` and returning a `bool`
      indicating match or mismatch with the desired pattern.

  Returns:
    A `set` of `building_blocks.ComputationBuildingBlock` instances
    representing the nodes in `tree` dependent on nodes matching `predicate`.
  """

  class _NodeSet:

    def __init__(self):
      self.mapping = {}

    def add(self, comp):
      self.mapping[id(comp)] = comp

    def to_set(self):
      return set(self.mapping.values())

  dependent_nodes = _NodeSet()

  def _are_children_in_dependent_set(comp, symbol_tree):
    """Checks if the dependencies of `comp` are present in `dependent_nodes`."""
    if isinstance(
        comp,
        (
            building_blocks.CompiledComputation,
            building_blocks.Data,
            building_blocks.Intrinsic,
            building_blocks.Literal,
            building_blocks.Placement,
        ),
    ):
      return False
    elif isinstance(comp, building_blocks.Lambda):
      return id(comp.result) in dependent_nodes.mapping
    elif isinstance(comp, building_blocks.Block):
      return (
          any(id(x[1]) in dependent_nodes.mapping for x in comp.locals)
          or id(comp.result) in dependent_nodes.mapping
      )
    elif isinstance(comp, building_blocks.Struct):
      return any(id(x) in dependent_nodes.mapping for x in comp)
    elif isinstance(comp, building_blocks.Selection):
      return id(comp.source) in dependent_nodes.mapping
    elif isinstance(comp, building_blocks.Call):
      return (
          id(comp.function) in dependent_nodes.mapping
          or id(comp.argument) in dependent_nodes.mapping
      )
    elif isinstance(comp, building_blocks.Reference):
      return _is_reference_dependent(comp, symbol_tree)

  def _is_reference_dependent(comp, symbol_tree):
    payload = symbol_tree.get_payload_with_name(comp.name)
    if payload is None:
      return False
    # The postorder traversal ensures that we process any
    # bindings before we process the reference to those bindings
    return id(payload.value) in dependent_nodes.mapping

  def _populate_dependent_set(comp, symbol_tree):
    """Populates `dependent_nodes` with all nodes dependent on `predicate`."""
    if predicate(comp):
      dependent_nodes.add(comp)
    elif _are_children_in_dependent_set(comp, symbol_tree):
      dependent_nodes.add(comp)
    return comp, False

  symbol_tree = transformation_utils.SymbolTree(
      transformation_utils.ReferenceCounter
  )
  transformation_utils.transform_postorder_with_symbol_bindings(
      tree, _populate_dependent_set, symbol_tree
  )
  return dependent_nodes.to_set()


def _extract_calls_with_fn_consuming_arg(
    tree: building_blocks.ComputationBuildingBlock,
    *,
    fn_predicate: _BuildingBlockPredicate,
    arg_predicate: _BuildingBlockPredicate,
) -> list[building_blocks.Call]:
  """Extracts calls depending on function and arg predicates.

  This function returns all calls in `tree` whose fns consume nodes matching
  `fn_predicate` and arguments consume nodes matching `arg_predicate`. This
  matching can be useful in checking that one type of function does not consume
  any nodes depending on another type of function in the body of `tree`.

  Args:
    tree: Instance of `building_blocks.ComputationBuildingBlock` to traverse.
    fn_predicate: Callable taking a building block and returning a boolean, to
      define the behavior of this function according to the semantics above.
    arg_predicate: Callable taking a building block and returning a boolean, to
      define the behavior of this function according to the semantics above.

  Returns:
    A list of `building_block.Calls` matching the description above.
  """
  nodes_dependent_on_arg_predicate = _extract_nodes_consuming(
      tree, arg_predicate
  )

  nodes_dependent_on_fn_predicate = _extract_nodes_consuming(tree, fn_predicate)

  instances = []

  for node in nodes_dependent_on_arg_predicate:
    if isinstance(node, building_blocks.Call):
      if (
          node.argument in nodes_dependent_on_arg_predicate
          and node.function in nodes_dependent_on_fn_predicate
      ):
        instances.append(node)
  return instances


def check_broadcast_not_dependent_on_aggregate(
    tree: building_blocks.ComputationBuildingBlock,
) -> None:
  """Raises if any broadcast in `tree` ingests the result of an aggregate.

  We explicitly check for this pattern since if it occurs, `tree` is not
  reducible to broadcast-map-aggregate form.


  Args:
    tree: Instance of `building_blocks.ComputationBuildingBlock` to check for
      the presence of a broadcast which ingests the result of an aggregate.

  Raises:
    ValueError: If a broadcast in `tree` consumes the result of an aggregate.
  """

  def aggregate_predicate(x):
    return (
        isinstance(x, building_blocks.Intrinsic)
        and x.intrinsic_def().aggregation_kind
    )

  def broadcast_predicate(x):
    return (
        isinstance(x, building_blocks.Intrinsic)
        and x.intrinsic_def().broadcast_kind
    )

  broadcast_dependent_examples = _extract_calls_with_fn_consuming_arg(
      tree, fn_predicate=broadcast_predicate, arg_predicate=aggregate_predicate
  )
  if broadcast_dependent_examples:
    raise ValueError(
        'Detected broadcast dependent on aggregate. Examples are: {}'.format(
            broadcast_dependent_examples
        )
    )


def check_aggregate_not_dependent_on_aggregate(
    tree: building_blocks.ComputationBuildingBlock,
) -> None:
  """Raises if any aggregation in `tree` ingests the result of an aggregate.

  We explicitly check for this pattern since if it occurs, `tree` is not
  reducible to `MergeableCompForm`.

  Args:
    tree: Instance of `building_blocks.ComputationBuildingBlock` to check for
      the presence of an aggregation which ingests the result of another
      aggregate.

  Raises:
    ValueError: If a broadcast in `tree` consumes the result of an aggregate.
  """

  def aggregate_predicate(x):
    return (
        isinstance(x, building_blocks.Intrinsic)
        and x.intrinsic_def().aggregation_kind
    )

  multiple_agg_dependent_examples = _extract_calls_with_fn_consuming_arg(
      tree, fn_predicate=aggregate_predicate, arg_predicate=aggregate_predicate
  )
  if multiple_agg_dependent_examples:
    raise ValueError(
        'Detected one aggregate dependent on another. Examples are: {}'.format(
            multiple_agg_dependent_examples
        )
    )


def check_contains_no_unbound_references(tree, excluding=None):
  """Checks that `tree` has no unbound references.

  Args:
    tree: Instance of `building_blocks.ComputationBuildingBlock` to view as an
      abstract syntax tree.
    excluding: A `string` or a collection of `string`s representing the names of
      references to exclude from the test.

  Raises:
    ValueError: If `comp` has unbound references.
  """
  if not contains_no_unbound_references(tree, excluding):
    raise ValueError(
        'The AST contains unbound references: {}.'.format(
            tree.formatted_representation()
        )
    )


def check_contains_no_new_unbound_references(old_tree, new_tree):
  """Checks that `new_tree` contains no unbound references not in `old_tree`."""
  old_unbound = transformation_utils.get_map_of_unbound_references(old_tree)[
      old_tree
  ]
  new_unbound = transformation_utils.get_map_of_unbound_references(new_tree)[
      new_tree
  ]
  diff = new_unbound - old_unbound
  if diff:
    raise ValueError(
        'Expected no new unbounded references. '
        f'Old tree:\n{old_tree}\nNew tree:\n{new_tree}\n'
        f'New unbound references: {diff}'
    )


def contains_called_intrinsic(tree, uri=None):
  """Tests if `tree` contains a called intrinsic for the given `uri`.

  Args:
    tree: A `building_blocks.ComputationBuildingBlock`.
    uri: An optional URI or list of URIs; the same as what is accepted by
      `building_block_analysis.is_called_intrinsic`.

  Returns:
    `True` if there is a called intrinsic in `tree` for the given `uri`,
    otherwise `False`.
  """
  predicate = lambda x: building_block_analysis.is_called_intrinsic(x, uri)
  return count(tree, predicate) > 0


def contains_no_unbound_references(
    tree: building_blocks.ComputationBuildingBlock,
    excluding: Optional[Union[str, Iterable[str]]] = None,
) -> bool:
  """Tests if all the references in `tree` are bound by `tree`.

  Args:
    tree: Instance of `building_blocks.ComputationBuildingBlock` to view as an
      abstract syntax tree.
    excluding: A `string` or a collection of `string`s representing the names of
      references to exclude from the test.

  Returns:
    `True` if there are no unbound references in `tree` excluding those
    specified by `excluding`, otherwise `False`.
  """
  if isinstance(excluding, str):
    excluding = [excluding]
  unbound_references = transformation_utils.get_map_of_unbound_references(tree)
  if excluding is not None:
    excluding = set(excluding)
    names = unbound_references[tree] - excluding
  else:
    names = unbound_references[tree]
  num_unbound_references = len(names)
  return num_unbound_references == 0


_DEFAULT_KIND_PREDICATE = lambda k: k is not None


def find_aggregations_in_tree(
    comp: building_blocks.ComputationBuildingBlock,
    kind_predicate: Callable[
        [intrinsic_defs.AggregationKind], bool
    ] = _DEFAULT_KIND_PREDICATE,
) -> list[building_blocks.Call]:
  """Finds aggregating calls with kind matching `kind_predicate` in `comp`.

  An "aggregating call" for the purpose of this function is a call to an
  intrinsic which takes values at CLIENT and materializes some result at
  SERVER.

  Args:
    comp: An AST to search.
    kind_predicate: A filter for kind of aggregation to search for.

  Returns:
    A list of child ASTs which are calls to aggregating intrinsics with kinds
    matching `aggregation_kind`.

  Raises:
    ValueError if `comp` contains a call whose target function cannot be
      identified. This may result from calls to references or other
      indirect structures.
  """
  aggregation_calls: list[building_blocks.Call] = []

  def record_intrinsic_calls(comp):
    """Identifies matching calls and adds them to `aggregation_calls`."""
    if not isinstance(comp, building_blocks.Call):
      return
    # Called lambdas will only trigger aggregation if they themselves contain
    # aggregation, which will be caught when the lambea itself is traversed.
    if isinstance(comp.function, building_blocks.Lambda):
      return
    # Aggregation cannot be occurring if the output type is not federated
    if not isinstance(
        comp.function.type_signature, computation_types.FunctionType
    ):
      return
    if not type_analysis.contains_federated_types(
        comp.function.type_signature.result
    ):
      return

    # We can't tell whether an arbitrary AST fragment results in an intrinsic
    # with a given URI, so we report an error in this case.
    if not isinstance(comp.function, building_blocks.Intrinsic):
      raise ValueError(
          'Cannot determine whether call contains aggregation: ' + str(comp)
      )

    # Aggregation with inputs that don't contain any tensors isn't interesting.
    #
    # NOTE: this is only applicable to intrinsic calls. Users can write their
    # own functions that internally materialize values at clients + aggregate
    # without taking any input tensors.
    #
    # This means that this check *must* come after the check above ensuring
    # that we're only talking about calls to `building_blocks.Intrinsic`s.
    if comp.argument is None or not type_analysis.contains_tensor_types(
        comp.argument.type_signature
    ):
      return

    if kind_predicate(comp.function.intrinsic_def().aggregation_kind):
      aggregation_calls.append(comp)

  visit_postorder(comp, record_intrinsic_calls)
  return aggregation_calls



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines classes/functions to manipulate the API context stack."""

from collections.abc import Iterator
import contextlib
import threading
from typing import Union

from federated_language.context_stack import context
from federated_language.context_stack import runtime_error_context


_Context = Union[context.AsyncContext, context.SyncContext]


class ContextStack(threading.local):
  """An thread-local context stack to run against."""

  def __init__(self, default_context: _Context):
    super().__init__()
    self._stack = [default_context]

  def set_default_context(self, ctx: _Context) -> None:
    """Installs `ctx` as the default context at the bottom of the stack."""
    self._stack[0] = ctx

  @property
  def current(self) -> _Context:
    """Returns the current context (one at the top of the context stack)."""
    return self._stack[-1]

  @contextlib.contextmanager
  def install(self, ctx: _Context) -> Iterator[_Context]:
    """A context manager that temporarily installs a new context on the stack.

    The installed context is placed at the top on the stack while in the context
    manager's scope, and remove from the stack upon exiting the scope. This
    method should only be used by the implementation code, and by the unit tests
    for dependency injection.

    Args:
      ctx: The context to temporarily install at the top of the context stack,
        an instance of `Context` defined in `context.py`.

    Yields:
      The installed context.

    Raises:
      TypeError: If `ctx` is not a valid instance of
        `federated_language.framework.AsyncContext` or
        `federated_language.framework.SyncContext`.
    """
    self._stack.append(ctx)
    try:
      yield ctx
    finally:
      self._stack.pop()


context_stack = ContextStack(runtime_error_context.RuntimeErrorContext())


def get_context_stack() -> ContextStack:
  """Returns the global context stack."""
  return context_stack


def set_default_context(ctx: _Context) -> None:
  """Installs `ctx` as the default context in the global context stack."""
  context_stack.set_default_context(ctx)


def set_no_default_context() -> None:
  context_stack.set_default_context(runtime_error_context.RuntimeErrorContext())



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines an abstract interface for things that possess type signatures."""

import abc

from federated_language.types import computation_types


class TypedObject(abc.ABC):
  """An abstract interface for things that possess type signatures."""

  @property
  @abc.abstractmethod
  def type_signature(self) -> computation_types.Type:
    """Returns the instance of `federated_language.Type` of this object."""
    raise NotImplementedError



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines interface for contexts which can bind symbols."""

import abc
from typing import Generic, TypeVar

from federated_language.context_stack import context


_Symbol = TypeVar('_Symbol')
_Reference = TypeVar('_Reference')


class SymbolBindingContext(
    context.SyncContext, abc.ABC, Generic[_Symbol, _Reference]
):
  """Interface for contexts which handle binding and tracking of references."""

  @abc.abstractmethod
  def bind_computation_to_reference(self, comp: _Symbol) -> _Reference:
    """Binds a computation to a symbol, returns a reference to this binding."""
    raise NotImplementedError

  @property
  @abc.abstractmethod
  def symbol_bindings(self) -> list[tuple[str, _Symbol]]:
    """Returns all symbols bound in this context."""
    raise NotImplementedError



# Copyright 2022, The TensorFlow Federated Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A federated platform implemented using native components."""

import asyncio
from collections.abc import Mapping
from typing import Optional, Union

from federated_language.common_libs import structure
from federated_language.computation import computation_base
from federated_language.context_stack import context
from federated_language.program import federated_context
from federated_language.program import structure_utils
from federated_language.program import value_reference
from federated_language.types import computation_types
from federated_language.types import placements
from federated_language.types import type_conversions
import tree


class NativeValueReference(value_reference.MaterializableValueReference):
  """A `federated_language.program.MaterializableValueReference` backed by a task."""

  def __init__(
      self,
      task: asyncio.Task,
      type_signature: value_reference.MaterializableTypeSignature,
  ):
    """Returns an initialized `federated_language.program.NativeValueReference`.

    Args:
      task: An `asyncio.Task` to run.
      type_signature: The `federated_language.Type` of this object.
    """
    self._task = task
    self._type_signature = type_signature

  @property
  def type_signature(
      self,
  ) -> value_reference.MaterializableTypeSignature:
    """The `federated_language.TensorType` of this object."""
    return self._type_signature

  async def get_value(self) -> value_reference.MaterializedValue:
    """Returns the referenced value as a numpy scalar or array."""
    return await self._task

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, NativeValueReference):
      return NotImplemented
    return (
        self._type_signature,
        self._task,
    ) == (
        other._type_signature,
        other._task,
    )


def _create_structure_of_references(
    task: asyncio.Task,
    type_signature: computation_types.Type,
) -> structure_utils.Structure[NativeValueReference]:
  """Returns a structure of `federated_language.program.NativeValueReference`s.

  Args:
    task: A task used to create the structure of
      `federated_language.program.NativeValueReference`s.
    type_signature: The `federated_language.Type` of the value returned by
      `task`; must contain only structures, server-placed values, or tensors.

  Raises:
    NotImplementedError: If `type_signature` contains an unexpected type.
  """
  if isinstance(type_signature, computation_types.StructType):

    def _get_container_cls(
        type_spec: computation_types.StructType,
    ) -> type[object]:
      container_cls = type_spec.python_container
      if container_cls is None:
        has_names = [name is not None for name, _ in type_spec.items()]
        if any(has_names):
          if not all(has_names):
            raise ValueError(
                'Expected `type_spec` to have either all named or unnamed'
                f' elements, found {type_spec}.'
            )
          container_cls = dict
        else:
          container_cls = list
      return container_cls

    async def _get_item(
        task: asyncio.Task, key: Union[str, int]
    ) -> value_reference.MaterializedValue:
      value = await task
      return value[key]

    elements = []
    for index, (name, element_type) in enumerate(type_signature.items()):
      container_cls = _get_container_cls(type_signature)
      if issubclass(container_cls, Mapping):
        key = name
      else:
        key = index
      element = _get_item(task, key)
      element_task = asyncio.create_task(element)
      element = _create_structure_of_references(element_task, element_type)
      elements.append(element)
    return type_conversions.to_structure_with_type(elements, type_signature)
  elif (
      isinstance(type_signature, computation_types.FederatedType)
      and type_signature.placement == placements.SERVER
  ):
    return _create_structure_of_references(task, type_signature.member)
  elif isinstance(type_signature, computation_types.SequenceType):
    return NativeValueReference(task, type_signature)
  elif isinstance(type_signature, computation_types.TensorType):
    return NativeValueReference(task, type_signature)
  else:
    raise NotImplementedError(f'Unexpected type found: {type_signature}.')


class NativeFederatedContext(federated_context.FederatedContext):
  """A `federated_language.program.FederatedContext` backed by a context."""

  def __init__(self, ctx: context.AsyncContext):
    """Returns an initialized `federated_language.program.NativeFederatedContext`.

    Args:
      ctx: An `federated_language.framework.AsyncContext`.
    """
    self._ctx = ctx

  def invoke(
      self,
      comp: computation_base.Computation,
      arg: Optional[federated_context.ComputationArg],
  ) -> structure_utils.Structure[NativeValueReference]:
    """Invokes the `comp` with the argument `arg`.

    Args:
      comp: The `federated_language.Computation` being invoked.
      arg: The optional argument of `comp`; server-placed values must be
        represented by `federated_language.program.MaterializableStructure`, and
        client-placed values must be represented by structures of values
        returned by a `federated_language.program.FederatedDataSourceIterator`.

    Returns:
      The result of invocation; a structure of
      `federated_language.program.MaterializableValueReference`.

    Raises:
      ValueError: If the result type of the invoked computation does not contain
      only structures, server-placed values, or tensors.
    Raises:
      ValueError: If the result type of `comp` does not contain only structures,
      server-placed values, or tensors.
    """
    result_type = comp.type_signature.result
    if not federated_context.contains_only_server_placed_data(result_type):
      raise ValueError(
          'Expected the result type of `comp` to contain only structures, '
          f'server-placed values, or tensors, found {result_type}.'
      )

    async def _invoke(
        ctx: context.AsyncContext,
        comp: computation_base.Computation,
        arg: value_reference.MaterializableStructure,
    ) -> value_reference.MaterializedStructure:
      if comp.type_signature.parameter is not None:

        def _to_python(obj):
          if isinstance(obj, structure.Struct):
            return structure.to_odict_or_tuple(obj)
          else:
            return None

        arg = tree.traverse(_to_python, arg)
        arg = await value_reference.materialize_value(arg)

      return await ctx.invoke(comp, arg)

    coro = _invoke(self._ctx, comp, arg)
    task = asyncio.create_task(coro)
    return _create_structure_of_references(task, result_type)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for Python functions, defuns, and other types of callables."""

from collections.abc import Mapping, Sequence
import inspect
import types
from typing import Optional

from federated_language.common_libs import structure
from federated_language.types import computation_types
from federated_language.types import type_conversions
from federated_language.types import typed_object


def _is_signature_compatible_with_types(
    signature: inspect.Signature, *args, **kwargs
) -> bool:
  """Determines if functions matching signature accept `args` and `kwargs`.

  Args:
    signature: An instance of `inspect.Signature` to verify agains the
      arguments.
    *args: Zero or more positional arguments, all of which must be instances of
      computation_types.Type or something convertible to it by
      computation_types.to_type().
    **kwargs: Zero or more keyword arguments, all of which must be instances of
      computation_types.Type or something convertible to it by
      computation_types.to_type().

  Returns:
    `True` or `False`, depending on the outcome of the test.

  Raises:
    TypeError: if the arguments are of the wrong computation_types.
  """
  try:
    bound_args = signature.bind(*args, **kwargs)
  except TypeError:
    return False

  # If we have no defaults then `bind` will have raised `TypeError` if the
  # signature was not compatible with *args and **kwargs.
  if all(
      p.default is inspect.Parameter.empty
      for p in signature.parameters.values()
  ):
    return True

  # Otherwise we need to check the defaults against the types that were given to
  # ensure they are compatible.
  for p in signature.parameters.values():
    if p.default is inspect.Parameter.empty or p.default is None:
      # No default value or optional.
      continue
    arg_value = bound_args.arguments.get(p.name, p.default)
    if arg_value is p.default:
      continue
    arg_type = computation_types.to_type(arg_value)
    default_type = type_conversions.infer_type(p.default)
    if not arg_type.is_assignable_from(default_type):
      return False
  return True


def _is_argument_struct(arg) -> bool:
  """Determines if 'arg' is interpretable as an argument struct.

  Args:
    arg: A value or type to test.

  Returns:
    True iff 'arg' is either a `Struct` in which all unnamed elements
    precede named ones, or a `StructType` with this property, or something
    that can be converted into the latter by computation_types.to_type().

  Raises:
    TypeError: If the argument is neither an `structure.Struct`,
      nor a type spec.
  """
  if isinstance(arg, structure.Struct):
    elements = structure.to_elements(arg)
  elif isinstance(arg, typed_object.TypedObject):
    return _is_argument_struct(arg.type_signature)
  else:
    if arg is not None:
      arg = computation_types.to_type(arg)
    if isinstance(arg, computation_types.StructType):
      elements = list(arg.items())
    else:
      return False
  max_unnamed = -1
  min_named = len(elements)
  for idx, element in enumerate(elements):
    if element[0]:
      min_named = min(min_named, idx)
    else:
      max_unnamed = idx
  return max_unnamed < min_named


def unpack_args_from_struct(
    struct_with_args,
) -> tuple[list[computation_types.Type], dict[str, computation_types.Type]]:
  """Extracts argument types from a struct.

  Args:
    struct_with_args: An instance of either an `struct.Struct` or
      computation_types.StructType` (or something convertible to it by
      computation_types.to_type()), on which _is_argument_struct() is True.

  Returns:
    A pair (args, kwargs) containing tuple elements from 'struct_with_args'.

  Raises:
    TypeError: if 'struct_with_args' is of a wrong type.
  """
  if not _is_argument_struct(struct_with_args):
    raise TypeError('Not an argument struct: {}.'.format(struct_with_args))
  if isinstance(struct_with_args, structure.Struct):
    elements = structure.to_elements(struct_with_args)
  elif isinstance(struct_with_args, typed_object.TypedObject):
    elements = []
    for index, (name, _) in enumerate(struct_with_args.type_signature.items()):  # pytype: disable=attribute-error
      if name is not None:
        elements.append((name, getattr(struct_with_args, name)))
      else:
        elements.append((None, struct_with_args[index]))
  else:
    struct_with_args = computation_types.to_type(struct_with_args)
    if not isinstance(struct_with_args, computation_types.StructType):
      raise ValueError(
          'Expected a `federated_language.StructType`, found'
          f' {struct_with_args}.'
      )
    elements = list(struct_with_args.items())
  args = []
  kwargs = {}
  for name, value in elements:
    if name is not None:
      kwargs[name] = value
    else:
      args.append(value)
  return args, kwargs


def pack_args_into_struct(
    args: Sequence[object], kwargs: Mapping[str, object], type_spec=None
) -> structure.Struct:
  """Packs positional and keyword arguments into a `Struct`.

  If 'type_spec' is not None, it must be a `StructType` or something that's
  convertible to it by computation_types.to_type(). The assignment of arguments
  to fields of the struct follows the same rule as during function calls. If
  'type_spec' is None, the positional arguments precede any of the keyword
  arguments, and the ordering of the keyword arguments matches the ordering in
  which they appear in kwargs. If the latter is an OrderedDict, the ordering
  will be preserved. On the other hand, if the latter is an ordinary unordered
  dict, the ordering is arbitrary.

  Args:
    args: Positional arguments.
    kwargs: Keyword arguments.
    type_spec: The optional type specification (either an instance of
      `computation_types.StructType` or something convertible to it), or None if
      there's no type. Used to drive the arrangements of args into fields of the
      constructed struct, as noted in the description.

  Returns:
    An struct containing all the arguments.

  Raises:
    TypeError: if the arguments are of the wrong computation_types.
  """
  if type_spec is not None:
    type_spec = computation_types.to_type(type_spec)
  if not type_spec:
    return structure.Struct(
        [(None, arg) for arg in args] + list(kwargs.items())
    )
  else:
    if not _is_argument_struct(type_spec):
      raise TypeError(
          'Parameter type {} does not have a structure of an argument struct, '
          'and cannot be populated from multiple positional and keyword '
          'arguments'.format(type_spec)
      )
    else:
      result_elements = []
      positions_used = set()
      keywords_used = set()
      for index, (name, elem_type) in enumerate(type_spec.items()):  # pytype: disable=attribute-error
        if index < len(args):
          # This argument is present in `args`.
          if name is not None and name in kwargs:
            raise TypeError('Argument `{}` specified twice.'.format(name))
          else:
            arg_value = args[index]
            result_elements.append((name, arg_value))
            positions_used.add(index)
        elif name is not None and name in kwargs:
          # This argument is present in `kwargs`.
          arg_value = kwargs[name]
          result_elements.append((name, arg_value))
          keywords_used.add(name)
        elif name:
          raise TypeError(f'Missing argument `{name}` of type {elem_type}.')
        else:
          raise TypeError(
              f'Missing argument of type {elem_type} at position {index}.'
          )
      positions_missing = set(range(len(args))).difference(positions_used)
      if positions_missing:
        raise TypeError(
            f'Positional arguments at {positions_missing} not used.'
        )
      keywords_missing = set(kwargs.keys()).difference(keywords_used)
      if keywords_missing:
        raise TypeError(f'Keyword arguments at {keywords_missing} not used.')
      return structure.Struct(result_elements)


def pack_args(
    parameter_type, args: Sequence[object], kwargs: Mapping[str, object]
):
  """Pack arguments into a single one that matches the given parameter type.

  The arguments may or may not be packed into a `Struct`, depending on the type
  of the parameter, and how many arguments are present.

  Args:
    parameter_type: The type of the single parameter expected by a computation,
      an instance of computation_types.Type or something convertible to it, or
      None if the computation is not expecting a parameter.
    args: Positional arguments of a call.
    kwargs: Keyword arguments of a call.

  Returns:
    A single value object of type that matches 'parameter_type' that contains
    all the arguments, or None if the 'parameter_type' is None.

  Raises:
    TypeError: if the args/kwargs do not match the given parameter type.
  """
  if parameter_type is None:
    # If there's no parameter type, there should be no args of any kind.
    if args or kwargs:
      raise TypeError('Was not expecting any arguments.')
    else:
      return None
  parameter_type = computation_types.to_type(parameter_type)
  if not args and not kwargs:
    raise TypeError(
        'Declared a parameter of type {}, but got no arguments.'.format(
            parameter_type
        )
    )
  single_positional_arg = len(args) == 1 and not kwargs
  if single_positional_arg:
    return args[0]
  if not isinstance(parameter_type, computation_types.StructType):
    # If not a `StructType`, a single positional argument is the only
    # supported call style.
    raise TypeError(
        'Parameter type {} is compatible only with a single positional '
        'argument, but found {} positional and {} keyword args.'.format(
            parameter_type, len(args), len(kwargs)
        )
    )
  if not _is_argument_struct(parameter_type):
    raise TypeError(
        'Parameter type {} does not have a structure of an argument '
        'struct, and cannot be populated from multiple positional and '
        'keyword arguments; please construct a struct before the '
        'call.'.format(parameter_type)
    )
  return pack_args_into_struct(args, kwargs, parameter_type)


def _infer_unpack_needed(
    fn: types.FunctionType,
    parameter_type: computation_types.Type,
    should_unpack: Optional[bool] = None,
) -> bool:
  """Returns whether parameter_type must be unpacked when calling fn.

  Args:
    fn: The function to be invoked.
    parameter_type: The type of the parameter bundle to be accepted by the
      returned callable, if any, or None if there's no parameter.
    should_unpack: Default or expected return value; None implies the inferred
      value should be returned. If either unpacking or packing could work, and
      should_unpack is not None, then should_unpack is returned.

  Returns:
    A `bool` indicating whether or not to unpack.
  """
  if should_unpack not in [True, False, None]:
    raise TypeError(
        'The unpack argument has an unexpected value {!r}.'.format(
            should_unpack
        )
    )
  unpack = should_unpack  # Default return value.
  signature = inspect.signature(fn)

  if parameter_type is None:
    if _is_signature_compatible_with_types(signature):
      if should_unpack:
        raise ValueError('Requested unpacking of a no-arg function.')
      return False
    else:
      raise TypeError(
          'The signature {} of the supplied function cannot be interpreted as '
          'a body of a no-parameter computation.'.format(signature)
      )

  unpack_required = not _is_signature_compatible_with_types(
      signature, parameter_type
  )
  if unpack_required and should_unpack is not None and not should_unpack:
    raise TypeError(
        "The supplied function '{}' with signature {} cannot accept a "
        "value of type '{}' as a single argument.".format(
            fn.__name__, signature, parameter_type
        )
    )
  if _is_argument_struct(parameter_type):
    arg_types, kwarg_types = unpack_args_from_struct(parameter_type)
    unpack_possible = _is_signature_compatible_with_types(
        signature, *arg_types, **kwarg_types
    )
  else:
    unpack_possible = False
  if not unpack_possible and should_unpack is not None and should_unpack:
    raise TypeError(
        'The supplied function with signature {} cannot accept a value of type '
        '{} as multiple positional and/or keyword arguments. That is, the '
        'argument cannot be unpacked, but unpacking was requested.'.format(
            signature, parameter_type
        )
    )
  if unpack_required and not unpack_possible:
    raise TypeError(
        'The supplied function "{}" with signature {} cannot accept a value of '
        'type {} as either a single argument or multiple positional and/or '
        'keyword arguments.'.format(fn.__name__, signature, parameter_type)
    )
  if not unpack_required and unpack_possible and should_unpack is None:
    # The supplied function could accept a value as either a single argument,
    # or as multiple positional and/or keyword arguments, and the caller did
    # not specify any preference, leaving ambiguity in how to handle the
    # mapping. We resolve the ambiguity by defaulting to capturing the entire
    # argument, as that's the behavior suggested as expected by the users.
    unpack = False

  if unpack is None:
    # Any ambiguity at this point has been resolved, so the following
    # condition holds and need only be verified in tests.
    assert unpack_required == unpack_possible, (
        unpack_required,
        unpack_possible,
    )
    unpack = unpack_possible

  return unpack


def wrap_as_zero_or_one_arg_callable(
    fn: types.FunctionType,
    parameter_type: Optional[computation_types.Type] = None,
    unpack: Optional[bool] = None,
):
  """Wraps around `fn` so it accepts up to one positional typed argument.

  This function helps to simplify dealing with functions and defuns that might
  have diverse and complex signatures, but that represent computations and as
  such, conceptually only accept a single parameter. The returned callable has
  a single positional parameter or no parameters. If it has one parameter, the
  parameter is expected to contain all arguments required by `fn` and matching
  the supplied parameter type signature bundled together into a `Struct`,
  if needed. The callable unpacks that structure, and passes all of
  its elements as positional or keyword-based arguments in the call to `fn`.

  Example usage:

    @tf.function
    def my_fn(x, y, z=10, name='bar', *p, **q):
      return x + y

    type_spec = (np.int32, np.int32)

    wrapped_fn = wrap_as_zero_or_one_arg_callable(my_fn, type_spec)

    arg = Struct([('x', 10), ('y', 20)])

    ... = wrapped_fn(arg)

  Args:
    fn: The underlying backend function or defun to invoke with the unpacked
      arguments.
    parameter_type: The type of the parameter bundle to be accepted by the
      returned callable, if any, or None if there's no parameter.
    unpack: Whether to break the parameter down into constituent parts and feed
      them as arguments to `fn` (True), leave the parameter as is and pass it to
      `fn` as a single unit (False), or allow it to be inferred from the
      signature of `fn` (None). In the latter case (None), if any ambiguity
      arises, an exception is thrown. If the parameter_type is None, this value
      has no effect, and is simply ignored.

  Returns:
    The zero- or one-argument callable that invokes `fn` with the unbundled
    arguments, as described above.

  Raises:
    TypeError: if arguments to this call are of the wrong types, or if the
      supplied 'parameter_type' is not compatible with `fn`.
  """
  signature = inspect.signature(fn)
  if parameter_type is None:
    if _is_signature_compatible_with_types(signature):
      # Deliberate wrapping to isolate the caller from `fn`, e.g., to prevent
      # the caller from mistakenly specifying args that match fn's defaults.
      return lambda: fn()  # pylint: disable=unnecessary-lambda
    else:
      raise TypeError(
          'The signature {} of the supplied function cannot be interpreted as '
          'a body of a no-parameter computation.'.format(signature)
      )
  else:
    parameter_type = computation_types.to_type(parameter_type)

    def _call(fn, parameter_type, arg, unpack):
      args, kwargs = unpack_arg(fn, parameter_type, arg, unpack)
      return fn(*args, **kwargs)

    # TODO: b/132888123 - Consider other options to avoid possible bugs here.
    try:
      (fn, parameter_type, unpack)
    except NameError as e:
      raise AssertionError('Args to be bound must be in scope.') from e
    return lambda arg: _call(fn, parameter_type, arg, unpack)


def _is_struct_with_py_container(value, type_spec):
  return isinstance(value, structure.Struct) and isinstance(
      type_spec, computation_types.StructWithPythonType
  )


def _unpack_arg(
    arg_types, kwarg_types, arg
) -> tuple[list[object], dict[str, object]]:
  """Unpacks 'arg' into an argument list based on types."""
  args = []
  for idx, expected_type in enumerate(arg_types):
    element_value = arg[idx]
    if isinstance(element_value, structure.Struct):
      element_value = type_conversions.type_to_py_container(
          element_value, expected_type
      )
    args.append(element_value)
  kwargs = {}
  for name, expected_type in kwarg_types.items():
    element_value = arg[name]
    if _is_struct_with_py_container(element_value, expected_type):
      element_value = type_conversions.type_to_py_container(
          element_value, expected_type
      )
    kwargs[name] = element_value
  return args, kwargs


def _ensure_arg_type(
    parameter_type, arg
) -> tuple[list[object], dict[str, object]]:
  """Ensures that `arg` matches `parameter_type` before returning it."""
  if _is_struct_with_py_container(arg, parameter_type):
    arg = type_conversions.type_to_py_container(arg, parameter_type)
  return [arg], {}


def unpack_arg(
    fn: types.FunctionType,
    parameter_type: Optional[computation_types.Type],
    arg,
    unpack: Optional[bool] = None,
) -> tuple[list[object], dict[str, object]]:
  """Converts values into arguments to `fn`.

  Args:
    fn: The function to unpack arguments for.
    parameter_type: The type of the parameter bundle to be accepted by the
      returned callable.
    arg: The argument to unpack.
    unpack: Whether to break the parameter down into constituent parts (`True`),
      leave the parameter as a single unit (False), or allow it to be inferred
      from the signature of `fn` (None). In the latter case (None), if any
      ambiguity arises, an exception is thrown.

  Returns:
    The unpacked arg.
  """
  if parameter_type is None:
    return [], {}

  if _infer_unpack_needed(fn, parameter_type, unpack):
    arg_types, kwarg_types = unpack_args_from_struct(parameter_type)
    return _unpack_arg(arg_types, kwarg_types, arg)
  else:
    return _ensure_arg_type(parameter_type, arg)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines classes/functions to manipulate the API context stack."""

from federated_language.context_stack import context


class RuntimeErrorContext(context.SyncContext):
  """A context that will fail when used to invoke a computation."""

  def invoke(self, comp, arg):
    del comp, arg  # Unused

    raise RuntimeError('No default context installed.')



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# limitations under the License.
"""A library of construction functions for computation type structures."""

from federated_language.types import computation_types


def reduction_op(
    result_type_spec: computation_types.Type,
    element_type_spec: computation_types.Type,
) -> computation_types.Type:
  """Returns the type of a reduction operator of the form `(<U,T> -> U)`.

  Args:
    result_type_spec: A `computation_types.Type`, the result of reduction (`U`).
    element_type_spec: A `computation_types.Type`, the type of elements to be
      reduced (`T`).

  Returns:
    The type of the corresponding reduction operator (`(<U,T> -> U)`).
  """
  return computation_types.FunctionType(
      computation_types.StructType([result_type_spec, element_type_spec]),
      result_type_spec,
  )


def unary_op(type_spec: computation_types.Type) -> computation_types.Type:
  """Returns the type of an unary operator that operates on `type_spec`.

  Args:
    type_spec: A `computation_types.Type`.

  Returns:
    The type of the corresponding unary operator.
  """
  return computation_types.FunctionType(type_spec, type_spec)


def binary_op(type_spec: computation_types.Type) -> computation_types.Type:
  """Returns the type of a binary operator that operates on `type_spec`.

  Args:
    type_spec: A `computation_types.Type`.

  Returns:
    The type of the corresponding binary operator.
  """
  return reduction_op(type_spec, type_spec)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permisions and
# limitations under the License.
"""A library of static analysis functions for computations."""

from collections.abc import Callable

from federated_language.compiler import building_block_analysis
from federated_language.compiler import building_blocks
from federated_language.compiler import intrinsic_defs
from federated_language.compiler import tree_analysis
from federated_language.computation import computation_impl


def computation_contains(
    computation: computation_impl.ConcreteComputation,
    predicate: Callable[[building_blocks.ComputationBuildingBlock], bool],
) -> bool:
  """Returns whether or not a building block in `tree` matches `predicate`."""
  building_block = computation.to_building_block()
  return tree_analysis.count(building_block, predicate) != 0


def computation_contains_secure_aggregation(
    computation: computation_impl.ConcreteComputation,
):
  """Determins if `computation` contains a secure aggregation call."""
  predicate = lambda x: building_block_analysis.is_called_aggregation(
      x,
      kind=intrinsic_defs.AggregationKind.SECURE,
  )
  return computation_contains(computation, predicate)


def computation_contains_unsecure_aggregation(
    computation: computation_impl.ConcreteComputation,
):
  """Determins if `computation` contains a secure aggregation call."""
  predicate = lambda x: building_block_analysis.is_called_aggregation(
      x,
      kind=intrinsic_defs.AggregationKind.DEFAULT,
  )
  return computation_contains(computation, predicate)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for testing context stacks."""

import asyncio
from collections.abc import Callable
import contextlib
import functools
from typing import Union

from federated_language.context_stack import context
from federated_language.context_stack import context_stack_impl

_Context = Union[context.AsyncContext, context.SyncContext]
_ContextFactory = Callable[[], _Context]


def with_context(context_fn: _ContextFactory):
  """Returns a decorator for running a test in a context.

  Args:
    context_fn: A `Callable` that constructs a
      `federated_language.framework.AsyncContext` or
      `federated_language.framework.SyncContext` to install before invoking the
      decorated function.
  """

  def decorator(fn):

    @contextlib.contextmanager
    def _install_context(context_fn: _ContextFactory):
      ctx = context_fn()
      with context_stack_impl.context_stack.install(ctx):
        yield

    if asyncio.iscoroutinefunction(fn):

      @functools.wraps(fn)
      async def wrapper(*args, **kwargs):
        with _install_context(context_fn):
          return await fn(*args, **kwargs)

    else:

      @functools.wraps(fn)
      def wrapper(*args, **kwargs):
        with _install_context(context_fn):
          return fn(*args, **kwargs)

    return wrapper

  return decorator



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Placement literals for use in computation and type definitions."""


class PlacementLiteral:
  """A representation of one of the globally recognized placement literals."""

  def __init__(
      self, name: str, uri: str, default_all_equal: bool, description: str
  ):
    self._name = name
    self._uri = uri
    self._default_all_equal = default_all_equal
    self._description = description

  @property
  def name(self) -> str:
    return self._name

  @property
  def uri(self) -> str:
    return self._uri

  @property
  def default_all_equal(self) -> bool:
    return self._default_all_equal

  def __doc__(self) -> str:
    return self._description

  def __str__(self) -> str:
    return self._name

  def __repr__(self) -> str:
    return "PlacementLiteral('{}')".format(self._uri)

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, PlacementLiteral):
      return NotImplemented
    return self._uri == other.uri

  def __hash__(self) -> int:
    return hash(self._uri)


CLIENTS = PlacementLiteral(
    'CLIENTS',
    'clients',
    default_all_equal=False,
    description='The collective of all client devices.',
)

SERVER = PlacementLiteral(
    'SERVER',
    'server',
    default_all_equal=True,
    description='The single top-level central coordinator.',
)


def uri_to_placement_literal(uri: str) -> PlacementLiteral:
  """Returns the placement literal that corresponds to the given URI.

  Args:
    uri: The URI of the placement.

  Returns:
    The placement literal.

  Raises:
    ValueError: if there is no known placement literal with such URI.
  """
  for literal in [CLIENTS, SERVER]:
    if uri == literal.uri:
      return literal
  raise ValueError(f'There is no known literal with uri `{uri}`.')



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for constructing decorators for functions and `tf.function`s."""

import collections
from collections.abc import Callable, Iterable, Mapping
import inspect
from typing import Optional

from federated_language.computation import computation_base
from federated_language.computation import polymorphic_computation
from federated_language.types import computation_types
from federated_language.types import type_conversions


def _parameters(fn) -> Mapping[str, inspect.Parameter]:
  return inspect.signature(fn).parameters


def _check_parameters(parameters: Mapping[str, inspect.Parameter]) -> None:
  """Ensure only non-varargs positional-or-keyword arguments."""
  for parameter in parameters.values():
    if parameter.default is not inspect.Parameter.empty:
      # We don't have a way to build defaults into the function's type.
      raise TypeError(
          'Default parameters are not supported. Found parameter '
          f'`{parameter.name}` with default value {parameter.default}'
      )
    if parameter.kind is inspect.Parameter.POSITIONAL_ONLY:
      # We don't have a way to encode positional-only into the function's type.
      raise TypeError(
          'Positional-only parameters are not supported. Found parameter '
          f'`{parameter.name}` which appears before a `/` entry.'
      )
    if parameter.kind is inspect.Parameter.KEYWORD_ONLY:
      # We don't have a way to encode keyword-only into the function's type.
      raise TypeError(
          'Keyword-only arguments are not supported. Found parameter '
          f'`{parameter.name}` which appears after a `*` or `*args` entry.'
      )
    if parameter.kind in (
        inspect.Parameter.VAR_POSITIONAL,
        inspect.Parameter.VAR_KEYWORD,
    ):
      # For concrete functions, we can't determine at tracing time which
      # arguments should be bundled into args vs. kwargs, since arguments can
      # be passed by position *or* by keyword at later call sites.
      raise TypeError(
          'Parameter varargs are not supported. Found varargs parameter '
          f'`{parameter.name}`.'
      )
    if parameter.kind is not inspect.Parameter.POSITIONAL_OR_KEYWORD:
      raise AssertionError(f'Unexpected parameter kind: {parameter.kind}')


def _wrap_polymorphic(
    fn,
    wrapper_fn: Callable[..., computation_base.Computation],
    infer_type_fn,
) -> polymorphic_computation.PolymorphicComputation:
  """Wraps `fn` in `wrapper_fn` at invocation time."""
  try:
    name = fn.__name__
  except AttributeError:
    name = None

  def _polymorphic_wrapper(
      parameter_type: computation_types.Type, unpack: Optional[bool]
  ):
    return wrapper_fn(fn, parameter_type, unpack=unpack, name=name)

  return polymorphic_computation.PolymorphicComputation(
      _polymorphic_wrapper, infer_type_fn
  )


def _wrap_concrete(
    fn,
    wrapper_fn: Callable[..., computation_base.Computation],
    parameter_type,
) -> computation_base.Computation:
  """Wraps `fn` in `wrapper_fn` given the provided `parameter_type`."""
  try:
    name = fn.__name__
  except AttributeError:
    name = None

  concrete_fn = wrapper_fn(fn, parameter_type, unpack=None, name=name)
  result_parameter_type = concrete_fn.type_signature.parameter
  if (
      result_parameter_type is not None
      and not result_parameter_type.is_equivalent_to(parameter_type)
  ):
    raise TypeError(
        'Expected a concrete function that takes parameter {}, got one '
        'that takes {}.'.format(
            str(parameter_type), str(concrete_fn.type_signature.parameter)
        )
    )
  return concrete_fn


def _parameter_type(
    parameters: Mapping[str, inspect.Parameter],
    parameter_types: tuple[computation_types.Type, ...],
) -> Optional[computation_types.Type]:
  """Bundle any user-provided parameter types into a single argument type."""
  parameter_names = parameters.keys()
  if not parameter_types and not parameters:
    return None
  if len(parameter_types) == 1:
    parameter_type = parameter_types[0]
    if parameter_type is None and not parameters:
      return None
    if len(parameters) == 1:
      return parameter_type
    # There is a single parameter type but multiple parameters.
    if not isinstance(parameter_type, computation_types.StructType) or len(
        parameter_type
    ) != len(parameters):
      raise TypeError(
          f'Function with {len(parameters)} parameters must have a parameter '
          'type with the same number of parameters. Found parameter type '
          f'{parameter_type}.'
      )
    name_list_from_types = [
        n for n, _ in parameter_type.items() if n is not None
    ]
    if name_list_from_types:
      if len(name_list_from_types) != len(parameter_type):
        raise TypeError(
            'Types with both named and unnamed fields cannot be unpacked into '
            f'argument lists. Found parameter type {parameter_type}.'
        )
      if set(name_list_from_types) != set(parameter_names):
        raise TypeError(
            'Function argument names must match field names of parameter type. '
            f'Found argument names {parameter_names}, which do not match '
            f'{name_list_from_types}, the top-level fields of the parameter '
            f'type {parameter_type}.'
        )
      # The provided parameter type has all named fields which exactly match
      # the names of the function's parameters.
      return parameter_type
    else:
      # The provided parameter type has no named fields. Apply the names from
      # the function parameters.
      parameter_types = (v for (_, v) in parameter_type.items())
      return computation_types.StructWithPythonType(
          list(zip(parameter_names, parameter_types)), collections.OrderedDict
      )
  elif len(parameters) == 1:
    # If there are multiple provided argument types but the function being
    # decorated only accepts a single argument, tuple the arguments together.
    return computation_types.StructType(parameter_types)
  if len(parameters) != len(parameter_types):
    raise TypeError(
        f'Function with {len(parameters)} parameters is '
        f'incompatible with provided argument types {parameter_types}.'
    )
  # The function has `n` parameters and `n` parameter types.
  # Zip them up into a structure using the names from the function as keys.
  return computation_types.StructWithPythonType(
      list(zip(parameter_names, parameter_types)), collections.OrderedDict
  )


def _wrap(
    fn,
    wrapper_fn: Callable[..., computation_base.Computation],
    parameter_types: tuple[computation_types.Type, ...],
    infer_type_fn: Callable[[object], computation_types.Type],
):
  """Wraps a possibly-polymorphic `fn` in `wrapper_fn`.

  If `parameter_type` is `None` and `fn` takes any arguments (even with default
  values), `fn` is inferred to be polymorphic and won't be passed to
  `wrapper_fn` until invocation time (when concrete parameter types are
  available).

  `wrapper_fn` must accept three positional arguments and one defaulted argument
  `name`:

  * `target_fn`, the Python function to be wrapped.

  * `parameter_types`, the user-provded list of parameter types.

  * `unpack`, an argument which will be passed on to
    `function_utils.wrap_as_zero_or_one_arg_callable` when wrapping `target_fn`.
    See that function for details.

  * Optional `name`, the name of the function that is being wrapped (only for
    debugging purposes).

  Args:
    fn: The function or tf.function to wrap as a computation.
    wrapper_fn: The Python callable that performs actual wrapping. The object to
      be returned by this function should be an instance of a
      `federated_language.framework.ConcreteComputation`.
    parameter_types: Types of any arguments to `fn`.
    infer_type_fn: ...

  Returns:
    Either the result of wrapping (an object that represents the computation),
    or a polymorphic callable that performs wrapping upon invocation based on
    argument types. The returned function still may accept multiple
    arguments (it has not yet had
    `function_uils.wrap_as_zero_or_one_arg_callable` applied to it).
  """
  parameters = _parameters(fn)
  # NOTE: many of the properties checked here are only necessary for
  # non-polymorphic computations whose type signatures must be resolved
  # prior to use. However, we continue to enforce these requirements even
  # in the polymorphic case in order to avoid creating an inconsistency.
  _check_parameters(parameters)

  if parameters and not parameter_types:
    # There is no type specification, and the function/tf.function declares
    # parameters. Create a polymorphic template.
    wrapped_fn = _wrap_polymorphic(fn, wrapper_fn, infer_type_fn)
  else:
    # Either we have a concrete parameter type, or this is no-arg function.
    parameter_type = _parameter_type(parameters, parameter_types)
    wrapped_fn = _wrap_concrete(fn, wrapper_fn, parameter_type)

  # When applying a decorator, the __doc__ attribute with the documentation
  # in triple-quotes is not automatically transferred from the function on
  wrapped_fn.__doc__ = getattr(fn, '__doc__', None)
  return wrapped_fn


def _is_function(obj: object) -> bool:
  # Type specifications are supported (i.e. objects that can be turned into a
  # `federated_language.Type`) as arguments to a computation decorator. In some
  # cases those type specifications (e.g. np.int32) are a `type`, making them
  # `Callable`, but they should not be treated as the function being decorated.
  if isinstance(obj, type):
    return False
  return isinstance(obj, Callable)


class ComputationReturnedNoneError(ValueError):
  """Error for computations which return `None` or do not return."""

  def __init__(self, fn):
    code = fn.__code__
    line_number = code.co_firstlineno
    filename = code.co_filename
    message = (
        f'The function defined on line {line_number} of file {filename} '
        "returned `None` (or didn't explicitly `return` at all), but "
        'computations must return some non-`None` value.'
    )
    super().__init__(message)


class ComputationWrapper:
  """A class for creating wrappers that convert functions into computations.

  This class builds upon the _wrap() function defined above, adding on
  functionality shared between the computation and `federated_computation`
  decorators. The shared functionality includes relating formal Python function
  parameters and call arguments to types, packing and unpacking arguments,
  verifying types, and support for polymorphism.

  Here's how one can use `ComputationWrapper` to construct a decorator/wrapper
  named `xyz`:

  ```python
  def my_wrapper_fn(target_fn, parameter_type, unpack, name=None):
    ...
  xyz = computation_wrapper.ComputationWrapper(my_wrapper_fn)
  ```

  The resulting `xyz` can then be used either as an `@xyz(...)` decorator or as
  a manual wrapping function: `wrapped_fn = xyz(my_func, ...)`. The latter
  method may be preferable when using functions from an external module or
  for wrapping an anonymous lambda.

  The decorator can be used in two ways:
  1. Invoked with a single positional argument specifying the types of the
     function's arguments (`@xyz(some_argument_type)`).
  2. Invoked with no arguments (`@xyz` or `@xyz()`). This is used for functions
     which take no arguments, or functions which are polymorphic (used with
     multiple different argument types).

  Here's how the decorator behaves in each case:

  If the user specifies a tuple type in an unbundled form (simply by listing the
  types of its constituents as separate arguments), the tuple type is formed on
  the user's behalf for convenience.

  1. When the decorator is invoked with positional arguments:

     ```python
     @xyz(('x', np.int32), ('y', np.int32))
     ```

     The decorator arguments must be instances of `types.Type`, or something
     convertible to it by `types.to_type()`. The arguments are interpreted as
     the specification of the parameter of the computation being constructed by
     the decorator. Since the formal parameter to computations is always a
     single argument, multiple arguments to the decorator will be packed into a
     tuple type. This means that the following two invocations behave the same:

     ```
     @xyz(('x', np.int32), ('y', np.int32)) # gets packed into the below
     @xyz((('x', np.int32), ('y', np.int32)))
     ```

     In the above example, the computation will accept as an argument a pair
     of integers named `x` and `y`.

     The function being decorated this way must declare at least one parameter.

     a. If the Python function declares only one parameter, that parameter will
        receive all arguments packed into a single value:

        ```python
        @xyz(('x', np.int32), ('y', np.int32))
        def my_comp(coord):
          ... # use `coord.x` and `coord.y`
        ```

     b. If the Python function declares multiple parameters, the computation's
        parameter type must be convertible to type
        `federated_language.StructType`
        (usually a list containing types or pairs of `(str, types.Type)`.

        ```python
        # With explicitly named parameters
        @xyz(('x', np.int32), ('y', np.int32))
        def my_comp(x, y):
          ... # use `x` and `y`

        # Without explicitly named parameters
        @xyz(np.int32, np.int32)
        def my_comp(x, y):
          ... # use `x` and `y`
        ```

        The number and order of parameters in the decorator arguments and the
        Python function must match. For named elements, the names in the
        decorator and the Python function must also match.

  2. When the decorator is specified without arguments (`@xyz` or `@xyz()`):

     a. If the Python function declares no parameters, the decorator constructs
        a no-parameter computation, as in the following example:

        ```python
        @xyz
        def my_comp():
          ...
        ```

     b. If the function does declare at least one parameter, it is treated as a
        polymorphic function that's instantiated in each concrete context in
        which it's used based on the types of its arguments. The decorator still
        handles the plumbing and parameter type inference.

        For example:

        ```python
        @xyz
        def my_comp(x, y):
          ...
        ```

        In this case, `my_comp` becomes a polymorphic callable, with the actual
        construction postponed. Suppose it's then used as follows, e.g., in an
        orchestration context:

        ```python
        my_comp(5.0, True)
        ```

        At the time of invocation, the decorator uses the information contained
        in the call arguments 5.0 and True to infer the computation's parameter
        type signature, and once the types have been determined, proceeds in
        exactly the same manner as already described in (1) above.

        It is important to note that the arguments of the invocation are not
        simply passed into the body of the Python function being decorated.
        The parameter type inference step is all that differs between the
        polymorphic case and case (1) above.

        Polymorphic functions are the only case where no constraints exist on
        the kinds of arguments that may be present: declaring default values,
        `*args` or `**kwargs`, and any combination of those are valid. The
        mapping is resolved at invocation time based on arguments of the call,
        as in the example below:

        ```python
        @xyz
        def my_comp(x, y=True, *args, **kwargs):
          ...

        my_comp(1, False, 2, 3, 'foo', name='bar')
        ```

        As with all polymorphic functions, no construction is actually performed
        until invocation, and at invocation time, the default parameter values
        are used alongside those actually used during the invocation to infer
        the computation's parameter type. The code that actually constructs the
        computation is oblivious to whether parameters of the Python function
        being decorated were driven by the default values, or by the arguments
        of the actual call.

        Note that the last argument to the function in the example above will
        be inferred as type `('name', str)`, not just `str`.

  For more examples of usage, see `computation_wrapper_test`.
  """

  def __init__(
      self,
      wrapper_fn: Callable[..., computation_base.Computation],
      to_type_fn: Callable[
          [object], computation_types.Type
      ] = computation_types.to_type,
      infer_type_fn: Callable[
          [object], computation_types.Type
      ] = type_conversions.infer_type,
  ):
    """Construct a new wrapper/decorator for the given wrapper callable.

    Args:
      wrapper_fn: The Python callable that performs actual wrapping (as in the
        specification of `_wrap`).
      to_type_fn: A `Callable` used to convert a backend-specific type
        specification to a `federated_language.Type`.
      infer_type_fn: A `Callable` used to convert a backend-specific values to a
        `federated_language.Type`.

    Raises:
      TypeError: if the arguments are of the wrong types.
    """
    self._wrapper_fn = wrapper_fn
    self._to_type_fn = to_type_fn
    self._infer_type_fn = infer_type_fn

  def __call__(self, *args, **kwargs):
    """Handles the different modes of usage of the decorator/wrapper.

    This method only acts as a frontend that allows this class to be used as a
    decorator or wrapper in a variety of ways. The actual wrapping is performed
    by the private method `_wrap`.

    Args:
      *args: Positional arguments for the computation decorator.
      **kwargs: Keyword arguments passed to individual computation strategies.

    Returns:
      Either a result of wrapping, or a callable that expects a function,
      method, or a tf.function and performs wrapping on it, depending on
      specific usage pattern.

    Raises:
      TypeError: if the arguments are of the wrong types.
    """

    def _to_types(
        objs: Iterable[object],
    ) -> tuple[Optional[computation_types.Type], ...]:
      result = []
      for obj in objs:
        if obj is not None:
          result.append(self._to_type_fn(obj))
        else:
          result.append(None)
      return tuple(result)

    if not args:
      # If invoked as a decorator, and with an empty argument list as "@xyz()"
      # applied to a function definition, expect the Python function being
      # decorated to be passed in the subsequent call, and potentially create
      # a polymorphic callable. The parameter type is unspecified.
      # Deliberate wrapping with a lambda to prevent the caller from being able
      # to accidentally specify parameter type as a second argument.
      # The tricky partial recursion is needed to inline the logic in the
      # "success" case below.
      provided_types = []
      return lambda fn: _wrap(
          fn, self._wrapper_fn, provided_types, self._infer_type_fn
      )
    elif _is_function(args[0]):
      # If the first argument on the list is a Python function, instance method,
      # or a tf.function, this is the one that's being wrapped. This is the case
      # of either a decorator invocation without arguments as "@xyz" applied to
      # a function definition, of an inline invocation as
      # `... = xyz(lambda....).`
      # Any of the following arguments, if present, are the arguments to the
      # wrapper that are to be interpreted as the type specification.
      fn = args[0]
      provided_types = _to_types(args[1:])
      return _wrap(fn, self._wrapper_fn, provided_types, self._infer_type_fn)
    else:
      provided_types = _to_types(args)
      return lambda fn: _wrap(
          fn, self._wrapper_fn, provided_types, self._infer_type_fn
      )



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for instrumenting code with timing and tracing data.

This module provides several functions for preserving trace context across
various boundaries, namely between asyncio and regular python code:

  * wrap_coroutine_in_trace_context wraps a coroutine such that it
    inherits the ambient trace context. It should be used when executing a
    coroutine that should inherit trace context from the current thread or
    task.
  * EventLoops should use the Task factory provided by
    propagate_trace_context_task_factory by calling
    `set_task_factory(propagate_trace_context_task_factory)`.
"""

import abc
import asyncio
from collections.abc import Generator
import contextlib
import functools
import inspect
import random
import sys
import threading
import time
from typing import Generic, Optional, TypeVar, Union

from absl import logging


class TracedSpan:
  """The trace was wrapping a non-function span.

  This value will be given back from `TracingProvider::span`'s first `yield`
  if the trace was being used to wrap a `span` rather than a whole function.
  """

  pass


class TracedFunctionReturned:
  """The traced function returned successfully.

  This value will be given back from `TracingProvider::span`'s first `yield`
  if the function being traced returned normally. The return value will be kept
  in the `value` field.
  """

  def __init__(self, value):
    self.value = value


class TracedFunctionThrew:
  """The traced function threw an exception.

  This value will be given back from `TracingProvider::span`'s first `yield`
  if the function being traced threw an exception.
  """

  def __init__(self, error_type, error_value, traceback):
    self.error_type = error_type
    self.error_value = error_value
    self.traceback = traceback


TraceResult = Union[TracedSpan, TracedFunctionReturned, TracedFunctionThrew]

T = TypeVar('T')


class TracingProvider(Generic[T], abc.ABC):
  """Abstract base class for tracers."""

  @abc.abstractmethod
  def span(
      self,
      scope: str,
      sub_scope: str,
      nonce: int,
      parent_span_yield: Optional[T],
      fn_args: Optional[tuple[object, ...]],
      fn_kwargs: Optional[dict[str, object]],
      trace_opts: dict[str, object],
  ) -> Generator[T, TraceResult, None]:
    """Create a new tracing span.

    Args:
      scope: String name of the scope, often the class name.
      sub_scope: String name of the sub-scope, often the function name.
      nonce: Number used to correlate tracing messages relating to the same
        function invocation.
      parent_span_yield: The value yielded by the most recently started (and not
        exited) call to `span` on this `TracingProvider` on the current
        `asyncio.Task` or thread (when running outside of an async context).
      fn_args: When this tracing provider wraps a function, this will be a tuple
        containing all of the non-keyword function arguments.
      fn_kwargs: When this tracing provider wraps a function, this will be a
        dict containing all of the keyword function arguments.
      trace_opts: User-provided options to the span constructor.
        `TracingProvider`s should ignore unknown options.

    Returns:
      A `Generator` which will be immediately started and run up until it
      yields for the first time. The value yielded by this `Generator`
      will be passed on to nested calls to `span`. When the spanned code ends,
      a `TraceResult` will be passed back through the `yield`.
    """
    raise NotImplementedError

  def wrap_rpc(
      self, parent_span_yield: Optional[T]
  ) -> contextlib.AbstractContextManager[None]:
    """Wrap an RPC call so that it can carry over the `parent_span_yield`."""
    del parent_span_yield
    return contextlib.nullcontext()

  def receive_rpc(self) -> Optional[T]:
    """Unpack `parent_span_yield` from the receiving end of an RPC."""
    return None


class LoggingTracingProvider(TracingProvider[None]):
  """Implements TracingProvider and outputs the results via logging.

  This implementation does not require storing additional trace context state,
  so most methods are no-ops.
  """

  def span(
      self,
      scope: str,
      sub_scope: str,
      nonce: int,
      parent_span_yield: Optional[None],
      fn_args: Optional[tuple[object, ...]],
      fn_kwargs: Optional[dict[str, object]],
      trace_opts: dict[str, object],
  ) -> Generator[None, TraceResult, None]:
    assert parent_span_yield is None
    del parent_span_yield, fn_args, fn_kwargs, trace_opts
    start_time = time.time()
    logging.debug('(%s) Entering %s.%s', nonce, scope, sub_scope)
    yield None
    logging.debug(
        '(%s) Exiting %s.%s. Elapsed time %f',
        nonce,
        scope,
        sub_scope,
        time.time() - start_time,
    )


_global_tracing_providers = [LoggingTracingProvider()]


def trace(fn=None, **trace_kwargs):
  """Delegates to the current global `TracingProvider`.

  Note that this function adds a layer of indirection so that the decoration
  happens when the method is executed. This is necessary so that the current
  TracingProvider is used.

  Args:
    fn: Function to decorate.
    **trace_kwargs: Tracing options. Supported options differ by tracing
      provider.

  Returns:
    Decorated instance of fn.
  """
  if fn is None:
    return functools.partial(trace, **trace_kwargs)

  scope, sub_scope = _func_to_class_and_method(fn)

  # NOTE: in a classic "what color is your function" situation,
  # we unfortunately have to duplicate the wrapping of the
  # underlying function in order to cover both the sync and async cases.
  if inspect.iscoroutinefunction(fn):

    @functools.wraps(fn)
    async def async_trace(*fn_args, **fn_kwargs):
      # Produce the span generator
      span_gen = _span_generator(
          scope, sub_scope, trace_kwargs, fn_args=fn_args, fn_kwargs=fn_kwargs
      )
      # Run up until the first yield
      next(span_gen)
      completed = False
      # Run the underlying function, recording the resulting value or exception
      # and passing it back to the span generator
      try:
        result = await fn(*fn_args, **fn_kwargs)
        completed = True
        try:
          span_gen.send(TracedFunctionReturned(result))
        except StopIteration:
          pass
        return result
      except:
        if not completed:
          error_type, error_value, traceback = sys.exc_info()
          try:
            span_gen.send(
                TracedFunctionThrew(error_type, error_value, traceback)
            )
          except StopIteration:
            pass
        raise

    return async_trace
  else:

    @functools.wraps(fn)
    def sync_trace(*fn_args, **fn_kwargs):
      span_gen = _span_generator(
          scope, sub_scope, trace_kwargs, fn_args=fn_args, fn_kwargs=fn_kwargs
      )
      next(span_gen)
      completed = False
      try:
        result = fn(*fn_args, **fn_kwargs)
        completed = True
        try:
          span_gen.send(TracedFunctionReturned(result))
        except StopIteration:
          pass
        return result
      except:
        if not completed:
          error_type, error_value, traceback = sys.exc_info()
          try:
            span_gen.send(
                TracedFunctionThrew(error_type, error_value, traceback)
            )
          except StopIteration:
            pass
        raise

    return sync_trace


# The code below manages the active "span yields" for a task or thread.
# Here's a quick summary of how that works.
#
# A "span yield" is a value `yield`ed by the `TracingProvider.span` function.
# The span yields for the current encompassing span need to be tracked so that
# they can be passed to new calls to `span` as the `parent_span_yield`.
#
# Typically, these would be tracked with a thread-local. However, async tasks
# can interleave on a single thread, so it makes more sense for them to track
# "task locals".
#
# `_current_span_yields` and `_set_span_yields` below handle the logic of
# tracking these spans. If we're in an async context, they'll read and write
# to the current async tasks, but fall back to using a thread local if we're
# in a synchronous context.

# A single yielded value for each currently-active TracingProvider.
SpanYields = list[Optional[object]]


class ThreadLocalSpanYields(threading.local):
  """The span set for the current thread.

  This is only used when outside of an async context.
  """

  def __init__(self):
    super().__init__()
    self._span_yields: Optional[SpanYields] = None

  def set(self, span_yields: Optional[SpanYields]):
    self._span_yields = span_yields

  def get(self) -> Optional[SpanYields]:
    return self._span_yields


_non_async_span_yields = ThreadLocalSpanYields()


def _current_task() -> Optional[asyncio.Task]:
  """Get the current running task, or `None` if no task is running."""
  # NOTE: `current_task` returns `None` if there is no current task, but it
  # throws if no currently running async loop.
  try:
    return asyncio.current_task()
  except RuntimeError:
    return None


def _current_span_yields() -> SpanYields:
  """Returns the current parent span yield list."""
  task = _current_task()
  if task is None:
    # There is no current task, so we're not running in an async context.
    # Grab the spans from the current thread.
    spans = _non_async_span_yields.get()
  else:
    spans = getattr(task, 'trace_span_yields', None)
  if spans is None:
    spans = [None for _ in range(len(_global_tracing_providers))]
  assert len(_global_tracing_providers) == len(spans)
  return spans


def _set_span_yields(span_yields: Optional[SpanYields]):
  """Sets the current parent span list."""
  task = _current_task()
  if task is None:
    # There is no current task, so we're not running in an async context.
    # Set the spans for the current thread.
    _non_async_span_yields.set(span_yields)
  else:
    setattr(task, 'trace_span_yields', span_yields)


@contextlib.contextmanager
def _with_span_yields(span_yields: Optional[SpanYields]):
  """Context manager which sets and unsets the current parent span list."""
  old_span_yields = _current_span_yields()
  _set_span_yields(span_yields)
  yield None
  _set_span_yields(old_span_yields)


@contextlib.contextmanager
def span(scope, sub_scope, **trace_opts):
  """Creates a `ContextManager` that wraps the code in question with a span."""
  span_gen = _span_generator(scope, sub_scope, trace_opts)
  next(span_gen)
  yield
  try:
    span_gen.send(TracedSpan())
  except StopIteration:
    pass


def _span_generator(
    scope, sub_scope, trace_opts, fn_args=None, fn_kwargs=None
) -> Generator[None, TraceResult, None]:
  """Wraps up all the `TracingProvider.span` generators into one."""
  # Create a nonce so that all of the traces from this span can be associated
  # with one another.
  nonce = random.randrange(1000000000)
  # Call `span` on all the global `TraceProvider`s and run it up until `yield`.
  span_generators = []
  new_span_yields: SpanYields = []
  for tp, parent_span_yield in zip(
      _global_tracing_providers, _current_span_yields()
  ):
    new_span_gen = tp.span(
        scope,
        sub_scope,
        nonce,
        parent_span_yield,
        fn_args,
        fn_kwargs,
        trace_opts,
    )
    new_span_yield = next(new_span_gen)
    span_generators.append(new_span_gen)
    new_span_yields.append(new_span_yield)
  # Set the values yielded by the `span` calls above to be the current span
  # yields, and yield so that the function can be run to completion.
  with _with_span_yields(new_span_yields):
    result = yield None
  # Send the result of the function to all of the generators so that they can
  # complete.
  for span_gen in reversed(span_generators):
    try:
      span_gen.send(result)
    except StopIteration:
      pass


def propagate_trace_context_task_factory(loop, coro):
  """Creates a new task on `loop` to run `coro`, inheriting current spans."""
  child_task = asyncio.tasks.Task(coro, loop=loop)
  trace_span_yields = _current_span_yields()
  setattr(child_task, 'trace_span_yields', trace_span_yields)
  return child_task


def wrap_coroutine_in_current_trace_context(coro):
  """Wraps the coroutine in the currently active span."""
  trace_span_yields = _current_span_yields()

  async def _wrapped():
    with _with_span_yields(trace_span_yields):
      return await coro

  return _wrapped()


@contextlib.contextmanager
def wrap_rpc_in_trace_context():
  """Attempts to record the trace context into the enclosed RPC call."""
  with contextlib.ExitStack() as stack:
    for tp, parent_span_yield in zip(
        _global_tracing_providers, _current_span_yields()
    ):
      stack.enter_context(tp.wrap_rpc(parent_span_yield))
    yield None


def set_tracing_providers(tracing_providers: list[TracingProvider]):
  """Set the global list of tracing providers, replacing any existing."""
  global _global_tracing_providers
  _global_tracing_providers = tracing_providers


def _func_to_class_and_method(fn) -> tuple[str, str]:
  """Returns the names of the function's class and method."""
  split = fn.__qualname__.split('.')
  if len(split) >= 2:
    class_name = split[-2]
    method_name = split[-1]
  else:
    module_name = fn.__module__
    class_name = module_name.split('.')[-1]
    method_name = fn.__name__
  return class_name, method_name



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ExecutorFactory interface and simple implementation."""

import abc
from collections.abc import MutableMapping

from federated_language.executor import executor_base
from federated_language.types import placements

CardinalitiesType = MutableMapping[placements.PlacementLiteral, int]


class ExecutorFactory(abc.ABC):
  """Interface defining executor factories.

  `ExecutorFactory` should be considered to own the executors it creates; it
  is responsible for their instantiation and management.

  `ExecutorFactory` exposes two methods, `create_executor` and
  `clean_up_executors`. There is a particular coupling between these two
  methods; any executor returned by `create_executor` should not be used
  after `clean_up_executors` has been called without reinitialization. That is,
  `create_executor` should be called again, and `ExecutorFactory` will ensure
  that the returned executor is safe for use.
  """

  @abc.abstractmethod
  def create_executor(
      self, cardinalities: CardinalitiesType
  ) -> executor_base.Executor:
    """Abstract method to construct instance of `executor_base.Executor`.

    `create_executor` must accept a dict mapping
    `placements.PlacementLiterals` to `ints`, and return an
    `executor_base.Executor`.

    Args:
      cardinalities: a dict mapping instances of `placements.PlacementLiteral`
        to ints, specifying the population size at each placement.

    Returns:
      Instance of `executor_base.Executor`.
    """
    pass

  @abc.abstractmethod
  def clean_up_executor(self, cardinalities: CardinalitiesType):
    """Releases any resources associated to the given cardinalities.

    Note that calling this method may invalidate the state of any executors
    which have previously been returned by the factory with the `cardinalities`
    argument ; `create_executor` should be called again if a new executor which
    is safe to use is desired.

    Args:
      cardinalities: The cardinalities of the executor whose state we wish to
        clear.
    """
    pass



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A library of transformation utilities."""

import abc
import collections
from collections.abc import Callable
import itertools
import operator
import typing
from typing import Optional

from federated_language.compiler import building_blocks


def transform_postorder(
    comp: building_blocks.ComputationBuildingBlock, transform
):
  """Traverses `comp` recursively postorder and replaces its constituents.

  For each element of `comp` viewed as an expression tree, the transformation
  `transform` is applied first to building blocks it is parameterized by, then
  the element itself. The transformation `transform` should act as an identity
  function on the kinds of elements (computation building blocks) it does not
  care to transform. This corresponds to a post-order traversal of the
  expression tree, i.e., parameters are always transformed left-to-right (in
  the order in which they are listed in building block constructors), then the
  parent is visited and transformed with the already-visited, and possibly
  transformed arguments in place.

  NOTE: In particular, in `Call(f,x)`, both `f` and `x` are arguments to `Call`.
  Therefore, `f` is transformed into `f'`, next `x` into `x'` and finally,
  `Call(f',x')` is transformed at the end.

  Args:
    comp: A `computation_building_block.ComputationBuildingBlock` to traverse
      and transform bottom-up.
    transform: The transformation to apply locally to each building block in
      `comp`. It is a Python function that accepts a building block at input,
      and should return a (building block, bool) tuple as output, where the
      building block is a `computation_building_block.ComputationBuildingBlock`
      representing either the original building block or a transformed building
      block and the bool is a flag indicating if the building block was modified
      as.

  Returns:
    The result of applying `transform` to parts of `comp` in a bottom-up
    fashion, along with a Boolean with the value `True` if `comp` was
    transformed and `False` if it was not.

  Raises:
    TypeError: If the arguments are of the wrong computation_types.
    NotImplementedError: If the argument is a kind of computation building block
      that is currently not recognized.
  """
  if isinstance(
      comp,
      (
          building_blocks.CompiledComputation,
          building_blocks.Data,
          building_blocks.Intrinsic,
          building_blocks.Literal,
          building_blocks.Placement,
          building_blocks.Reference,
      ),
  ):
    return transform(comp)
  elif isinstance(comp, building_blocks.Selection):
    source, source_modified = transform_postorder(comp.source, transform)
    if source_modified:
      comp = building_blocks.Selection(source, comp.name, comp.index)
    comp, comp_modified = transform(comp)
    return comp, comp_modified or source_modified
  elif isinstance(comp, building_blocks.Struct):
    elements = []
    elements_modified = False
    for key, value in comp.items():
      value, value_modified = transform_postorder(value, transform)
      elements.append((key, value))
      elements_modified = elements_modified or value_modified
    if elements_modified:
      comp = building_blocks.Struct(
          elements, container_type=comp.type_signature.python_container
      )
    comp, comp_modified = transform(comp)
    return comp, comp_modified or elements_modified
  elif isinstance(comp, building_blocks.Call):
    fn, fn_modified = transform_postorder(comp.function, transform)
    if comp.argument is not None:
      arg, arg_modified = transform_postorder(comp.argument, transform)
    else:
      arg, arg_modified = (None, False)
    if fn_modified or arg_modified:
      comp = building_blocks.Call(fn, arg)
    comp, comp_modified = transform(comp)
    return comp, comp_modified or fn_modified or arg_modified
  elif isinstance(comp, building_blocks.Lambda):
    result, result_modified = transform_postorder(comp.result, transform)
    if result_modified:
      comp = building_blocks.Lambda(
          comp.parameter_name, comp.parameter_type, result
      )
    comp, comp_modified = transform(comp)
    return comp, comp_modified or result_modified
  elif isinstance(comp, building_blocks.Block):
    variables = []
    variables_modified = False
    for key, value in comp.locals:
      value, value_modified = transform_postorder(value, transform)
      variables.append((key, value))
      variables_modified = variables_modified or value_modified
    result, result_modified = transform_postorder(comp.result, transform)
    if variables_modified or result_modified:
      comp = building_blocks.Block(variables, result)
    comp, comp_modified = transform(comp)
    return comp, comp_modified or variables_modified or result_modified
  else:
    raise NotImplementedError(f'Unexpected building block found: {type(comp)}.')


TransformReturnType = tuple[building_blocks.ComputationBuildingBlock, bool]


def transform_preorder(
    comp: building_blocks.ComputationBuildingBlock,
    transform: Callable[
        [building_blocks.ComputationBuildingBlock], TransformReturnType
    ],
) -> TransformReturnType:
  """Walks the AST of `comp` preorder, calling `transform` on the way down.

  Notice that this function will stop walking the tree when its transform
  function modifies a node; this is to prevent the caller from unexpectedly
  kicking off an infinite recursion. For this purpose the transform function
  must identify when it has transformed the structure of a building block; if
  the structure of the building block is modified but `False` is returned as
  the second element of the tuple returned by `transform`, `transform_preorder`
  may result in an infinite recursion.

  Args:
    comp: Instance of `building_blocks.ComputationBuildingBlock` to be
      transformed in a preorder fashion.
    transform: Transform function to be applied to the nodes of `comp`. Must
      return a two-tuple whose first element is a
      `building_blocks.ComputationBuildingBlock` and whose second element is a
      Boolean. If the computation which is passed to `comp` is returned in a
      modified state, must return `True` for the second element. This Boolean
      controls whether or not to stop traversing the tree under `comp`; if this
      Bool is `True`, `transform_preorder` will not traverse this subtree.

  Returns:
    A two-tuple, whose first element is modified version of `comp`, and
    whose second element is a Boolean indicating whether `comp` was transformed
    during the walk.
  """

  inner_comp, modified = transform(comp)
  if modified:
    return inner_comp, modified
  if isinstance(
      inner_comp,
      (
          building_blocks.CompiledComputation,
          building_blocks.Data,
          building_blocks.Intrinsic,
          building_blocks.Literal,
          building_blocks.Placement,
          building_blocks.Reference,
      ),
  ):
    return inner_comp, modified
  elif isinstance(inner_comp, building_blocks.Lambda):
    transformed_result, result_modified = transform_preorder(
        inner_comp.result, transform
    )
    if not (modified or result_modified):
      return inner_comp, False
    return (
        building_blocks.Lambda(
            inner_comp.parameter_name,
            inner_comp.parameter_type,
            transformed_result,
        ),
        True,
    )
  elif isinstance(inner_comp, building_blocks.Struct):
    elements_modified = False
    elements = []
    for name, val in inner_comp.items():
      result, result_modified = transform_preorder(val, transform)
      elements_modified = elements_modified or result_modified
      elements.append((name, result))
    if not (modified or elements_modified):
      return inner_comp, False
    return building_blocks.Struct(elements), True
  elif isinstance(inner_comp, building_blocks.Selection):
    transformed_source, source_modified = transform_preorder(
        inner_comp.source, transform
    )
    if not (modified or source_modified):
      return inner_comp, False
    return (
        building_blocks.Selection(
            transformed_source, inner_comp.name, inner_comp.index
        ),
        True,
    )
  elif isinstance(inner_comp, building_blocks.Call):
    transformed_fn, fn_modified = transform_preorder(
        inner_comp.function, transform
    )
    if inner_comp.argument is not None:
      transformed_arg, arg_modified = transform_preorder(
          inner_comp.argument, transform
      )
    else:
      transformed_arg = None
      arg_modified = False
    if not (modified or fn_modified or arg_modified):
      return inner_comp, False
    return building_blocks.Call(transformed_fn, transformed_arg), True
  elif isinstance(inner_comp, building_blocks.Block):
    transformed_variables = []
    values_modified = False
    for key, value in inner_comp.locals:
      transformed_value, value_modified = transform_preorder(value, transform)
      transformed_variables.append((key, transformed_value))
      values_modified = values_modified or value_modified
    transformed_result, result_modified = transform_preorder(
        inner_comp.result, transform
    )
    if not (modified or values_modified or result_modified):
      return inner_comp, False
    return (
        building_blocks.Block(transformed_variables, transformed_result),
        True,
    )
  else:
    raise NotImplementedError(
        f'Unexpected building block found: {type(inner_comp)}.'
    )


def transform_postorder_with_symbol_bindings(
    comp: building_blocks.ComputationBuildingBlock,
    transform: Callable[
        [building_blocks.ComputationBuildingBlock, 'SymbolTree'],
        building_blocks.ComputationBuildingBlock,
    ],
    symbol_tree: 'SymbolTree',
):
  """Uses symbol binding hooks to execute transformations.

  `transform_postorder_with_symbol_bindings` hooks into the preorder traversal
  that is defined by walking down the tree to its leaves, using
  the variable bindings along this path to push information onto
  the given `SymbolTree`. Once we hit the leaves, we walk back up the
  tree in a postorder fashion, calling `transform` as we go.

  The transformations `transform_postorder_with_symbol_bindings` executes are
  therefore stateful in some sense. Here 'stateful' means that a transformation
  executed on a given AST node in general depends on not only the node itself
  or its immediate vicinity; possibly there is some global information on which
  this transformation depends. `transform_postorder_with_symbol_bindings` is
  functional 'from AST to AST' (where `comp` represents the root of an AST) but
  not 'from node to node'.

  One important fact to note: there are recursion invariants that
  `transform_postorder_with_symbol_bindings` uses the `SymbolTree` data
  structure to enforce. In particular, within a `transform` call the following
  invariants hold:

  * `symbol_tree.update_payload_with_name` with an argument `name` will call
    `update` on the `BoundVariableTracker` in `symbol_tree` which tracks the
    value of `ref` active in the current lexical scope. Will raise a
    `NameError` if none exists.

  * `symbol_tree.get_payload_with_name` with a string argument `name` will
    return the `BoundVariableTracker` instance from `symbol_tree` which
    corresponds to the computation bound to the variable `name` in the current
    lexical scope. Will raise a `NameError` if none exists.

  These recursion invariants are enforced by the framework, and should be
  relied on when designing new transformations that depend on variable
  bindings.

  Args:
    comp: Instance of `building_blocks.ComputationBuildingBlock` to read
      information from or transform.
    transform: Python function accepting `comp` and `symbol_tree` arguments and
      returning `transformed_comp`.
    symbol_tree: Instance of `SymbolTree`, the data structure into which we may
      read information about variable bindings, and from which we may read.

  Returns:
    Returns a possibly modified version of `comp`, an instance
    of `building_blocks.ComputationBuildingBlock`, along with a
    Boolean with the value `True` if `comp` was transformed and `False` if it
    was not.
  """
  identifier_seq = itertools.count(start=1)

  def _transform_postorder_with_symbol_bindings_switch(
      comp, transform, ctxt_tree, identifier_sequence
  ):
    """Recursive helper function delegated to after binding comp_id sequence."""
    if isinstance(
        comp,
        (
            building_blocks.CompiledComputation,
            building_blocks.Data,
            building_blocks.Intrinsic,
            building_blocks.Literal,
            building_blocks.Placement,
            building_blocks.Reference,
        ),
    ):
      return _traverse_leaf(comp, transform, ctxt_tree, identifier_sequence)
    elif isinstance(comp, building_blocks.Selection):
      return _traverse_selection(
          comp, transform, ctxt_tree, identifier_sequence
      )
    elif isinstance(comp, building_blocks.Struct):
      return _traverse_tuple(comp, transform, ctxt_tree, identifier_sequence)
    elif isinstance(comp, building_blocks.Call):
      return _traverse_call(comp, transform, ctxt_tree, identifier_sequence)
    elif isinstance(comp, building_blocks.Lambda):
      return _traverse_lambda(comp, transform, ctxt_tree, identifier_sequence)
    elif isinstance(comp, building_blocks.Block):
      return _traverse_block(comp, transform, ctxt_tree, identifier_sequence)
    else:
      raise NotImplementedError(
          f'Unexpected building block found: {type(comp)}.'
      )

  def _traverse_leaf(comp, transform, context_tree, identifier_seq):
    """Helper function holding traversal logic for leaf nodes."""
    _ = next(identifier_seq)
    return transform(comp, context_tree)

  def _traverse_selection(comp, transform, context_tree, identifier_seq):
    """Helper function holding traversal logic for selection nodes."""
    _ = next(identifier_seq)
    source, source_modified = _transform_postorder_with_symbol_bindings_switch(
        comp.source, transform, context_tree, identifier_seq
    )
    if source_modified:
      # Normalize selection to index based on the type signature of the
      # original source. The new source may not have names present.
      if comp.index is not None:
        index = comp.index
      else:
        names = [n for n, _ in comp.source.type_signature.items()]
        index = names.index(comp.name)
      comp = building_blocks.Selection(source, index=index)
    comp, comp_modified = transform(comp, context_tree)
    return comp, comp_modified or source_modified

  def _traverse_tuple(comp, transform, context_tree, identifier_seq):
    """Helper function holding traversal logic for tuple nodes."""
    _ = next(identifier_seq)
    elements = []
    elements_modified = False
    for key, value in comp.items():
      value, value_modified = _transform_postorder_with_symbol_bindings_switch(
          value, transform, context_tree, identifier_seq
      )
      elements.append((key, value))
      elements_modified = elements_modified or value_modified
    if elements_modified:
      comp = building_blocks.Struct(elements)
    comp, comp_modified = transform(comp, context_tree)
    return comp, comp_modified or elements_modified

  def _traverse_call(comp, transform, context_tree, identifier_seq):
    """Helper function holding traversal logic for call nodes."""
    _ = next(identifier_seq)
    fn, fn_modified = _transform_postorder_with_symbol_bindings_switch(
        comp.function, transform, context_tree, identifier_seq
    )
    if comp.argument is not None:
      arg, arg_modified = _transform_postorder_with_symbol_bindings_switch(
          comp.argument, transform, context_tree, identifier_seq
      )
    else:
      arg, arg_modified = (None, False)
    if fn_modified or arg_modified:
      comp = building_blocks.Call(fn, arg)
    comp, comp_modified = transform(comp, context_tree)
    return comp, comp_modified or fn_modified or arg_modified

  def _traverse_lambda(comp, transform, context_tree, identifier_seq):
    """Helper function holding traversal logic for lambda nodes."""
    comp_id = next(identifier_seq)
    context_tree.drop_scope_down(comp_id)
    context_tree.ingest_variable_binding(name=comp.parameter_name, value=None)
    result, result_modified = _transform_postorder_with_symbol_bindings_switch(
        comp.result, transform, context_tree, identifier_seq
    )
    context_tree.walk_to_scope_beginning()
    if result_modified:
      comp = building_blocks.Lambda(
          comp.parameter_name, comp.parameter_type, result
      )
    comp, comp_modified = transform(comp, context_tree)
    context_tree.pop_scope_up()
    return comp, comp_modified or result_modified

  def _traverse_block(comp, transform, context_tree, identifier_seq):
    """Helper function holding traversal logic for block nodes."""
    comp_id = next(identifier_seq)
    context_tree.drop_scope_down(comp_id)
    variables = []
    variables_modified = False
    for key, value in comp.locals:
      value, value_modified = _transform_postorder_with_symbol_bindings_switch(
          value, transform, context_tree, identifier_seq
      )
      context_tree.ingest_variable_binding(name=key, value=value)
      variables.append((key, value))
      variables_modified = variables_modified or value_modified
    result, result_modified = _transform_postorder_with_symbol_bindings_switch(
        comp.result, transform, context_tree, identifier_seq
    )
    context_tree.walk_to_scope_beginning()
    if variables_modified or result_modified:
      comp = building_blocks.Block(variables, result)
    comp, comp_modified = transform(comp, context_tree)
    context_tree.pop_scope_up()
    return comp, comp_modified or variables_modified or result_modified

  return _transform_postorder_with_symbol_bindings_switch(
      comp, transform, symbol_tree, identifier_seq
  )


class BoundVariableTracker(abc.ABC):
  """Abstract class representing a mutable variable binding."""

  def __init__(
      self, name: str, value: Optional[building_blocks.ComputationBuildingBlock]
  ):
    """Initializes `BoundVariableTracker`.

    The initializer is likely to be overwritten by subclasses in order to
    attach more state to the `BoundVariableTracker`. Each of them must
    satisfy the same interface, however. This is simply because the
    `BoundVariableTracker` represents a variable binding in an AST;
    no more information is avaiable to it than the `name`-`value` pair
    being bound together.

    Args:
      name: String name of variable to be bound.
      value: Value to bind to this name. Can be instance of
        `building_blocks.ComputationBuildingBlock` if this
        `BoundVariableTracker` represents a concrete binding to a variable (e.g.
        in a block locals declaration), or `None`, if this
        `BoundVariableTracker` represents merely a variable declaration (e.g. in
        a lambda).
    """
    self.name = name
    self.value = value

  def update(self, value=None):
    """Defines the way information is read into this node.

    Defaults to no-op.

    Args:
      value: Similar to `value` argument in initializer.
    """
    del value  # Unused

  @abc.abstractmethod
  def __str__(self):
    """Abstract string method required as context tree will delegate."""

  def __eq__(self, other):
    """Base class equality checks names and values equal."""
    if self is other:
      return True
    if not isinstance(other, BoundVariableTracker):
      return NotImplemented
    if self.name != other.name:
      return False
    if isinstance(
        self.value, building_blocks.ComputationBuildingBlock
    ) and isinstance(other.value, building_blocks.ComputationBuildingBlock):
      return (
          self.value.compact_representation()
          == other.value.compact_representation()
          and self.value.type_signature.is_equivalent_to(
              other.value.type_signature
          )
      )
    return self.value is other.value

  def __ne__(self, other):
    """Implementing __ne__ to enforce in Python2 the Python3 standard."""
    return not self == other


class _BeginScopePointer(BoundVariableTracker):
  """Sentinel representing the beginning of a scope defined by an AST node."""

  def __init__(self, name=None, value=None):
    if name is not None or value is not None:
      raise ValueError(
          "Please don't pass a name or value to "
          '_BeginScopePointer; it will simply be ignored.'
      )
    super().__init__('BeginScope', None)

  def update(self, value=None):
    del value  # Unused.
    raise RuntimeError("We shouldn't be trying to update the outer context.")

  def __str__(self):
    return self.name

  def __eq__(self, other):
    """Returns `True` iff `other` is also a `_BeginScopePointer`.

    Args:
      other: Value for equality comparison.

    Returns:
      Returns true iff `other` is also an instance of `_BeginScopePointer`.
    """
    # Using explicit type comparisons here to prevent a subclass from passing.
    # pylint: disable=unidiomatic-typecheck
    return type(other) is _BeginScopePointer
    # pylint: enable=unidiomatic-typecheck


class SymbolTree:
  """Data structure to hold variable bindings as we walk an AST.

  `SymbolTree` is designed to be constructed and mutatated as we traverse an
  AST, maintaining a pointer to an active node representing the variable
  bindings we currently have available as we walk the AST.

  `SymbolTree` is a hierarchical tree-like data structure. Its levels
  correspond to nodes in the AST it is tracking, meaning that walking into
  or out of a scope-defining node (a block or lambda) corresponds to
  moving up or down a level in the `SymbolTree`. Block constructs (a.k.a.
  the let statement) binds variables sequentially, and this sequential binding
  corresponds to variables bound at the same level of the `SymbolTree`.

  Each instance of the node class can be used at most once in the symbol tree,
  as checked by memory location. This disallows circular tree structures that
  could cause an infinite loop in recursive equality testing or printing.
  """

  def __init__(self, payload_type: type[BoundVariableTracker]):
    """Initializes `SymbolTree` with its payload type.

    Args:
      payload_type: Class which subclasses BoundVariableTracker; the type of
        payloads to be constructed and held in this SymbolTree.
    """
    initial_node = SequentialBindingNode(_BeginScopePointer())
    self.active_node = initial_node
    self.payload_type = payload_type
    self._node_ids = {id(initial_node): 1}

  def get_payload_with_name(self, name: str) -> Optional[BoundVariableTracker]:
    """Returns payload corresponding to `name` in active variable bindings.

    Note that this method obeys `dict.get`-like semantics; instead of raising
    when asked to address an unbound name, it simply returns `None`.

    Args:
      name: String name to find in currently active context.

    Returns:
      Returns instance of `BoundVariableTracker` corresponding to `name`
      in context represented by `active_comp`, or `None` if the requested
      name is unbound in the current context.
    """
    comp = typing.cast(SequentialBindingNode, self.active_node)
    while comp.parent is not None or comp.older_sibling is not None:
      if name == comp.payload.name:
        return comp.payload
      if comp.older_sibling is not None:
        comp = comp.older_sibling
      elif comp.parent is not None:
        comp = comp.parent
    return None

  def get_higher_payloads_with_value(self, value, equal_fn=None):
    """Returns payloads above `active_node` whose `value` is equal to `value`.

    Args:
      value: The value to test.
      equal_fn: The optional function to use to determine equality, if `None` is
        specified `operator.is_` is used.
    """
    payloads = []
    if equal_fn is None:
      equal_fn = operator.is_
    node = typing.cast(SequentialBindingNode, self.active_node)
    while node.parent is not None or node.older_sibling is not None:
      if node.payload.value is not None and equal_fn(value, node.payload.value):
        payloads.append(node.payload)
      if node.older_sibling is not None:
        node = node.older_sibling
      elif node.parent is not None:
        node = node.parent
    return payloads

  def update_payload_with_name(self, name: str):
    """Calls `update` if `name` is found among the available symbols.

    If there is no such available symbol, simply does nothing.

    Args:
      name: A string; generally, this is the variable a walker has encountered
        in a AST, and which it is relying on `SymbolTable` to address correctly.

    Raises:
      ValueError: If `name` is not found among the bound names currently
        available in `self`.
    """
    comp = typing.cast(SequentialBindingNode, self.active_node)
    while comp.parent is not None or comp.older_sibling is not None:
      if name == comp.payload.name:
        comp.payload.update(name)
        return None
      if comp.older_sibling is not None:
        comp = comp.older_sibling
      elif comp.parent is not None:
        comp = comp.parent
    raise ValueError(
        "The name '{}' is not available in '{}'.".format(name, self)
    )

  def walk_to_scope_beginning(self):
    """Walks `active_node` back to the sentinel node beginning current scope.

    `walk_to_scope_beginning` resolves the issue of scope at a node which
    introduces scope in the following manner: each of these nodes (for instance,
    a `building_blocks.Lambda`) corresponds to a sentinel value of
    the `_BeginScopePointer` class, ensuring that these nodes do not have access
    to
    scope that is technically not available to them. That is, we conceptualize
    the node corresponding to `(x -> x)` as existing in the scope outside of the
    binding of `x`, and therefore is unable to reference `x`. However, these
    nodes can walk down their variable declarations via
    `walk_down_one_variable_binding` in order to inspect these declarations and
    perhaps execute some logic based on them.
    """
    scope_sentinel = _BeginScopePointer()
    self.active_node = typing.cast(SequentialBindingNode, self.active_node)
    assert self.active_node is not None
    while self.active_node.payload != scope_sentinel:
      self.active_node = self.active_node.older_sibling
      assert self.active_node is not None

  def pop_scope_up(self):
    """Moves `active_node` up one level in the `SymbolTree`.

    Raises:
      Raises ValueError if we are already at the highest level.
    """
    self.walk_to_scope_beginning()
    self.active_node = typing.cast(SequentialBindingNode, self.active_node)
    if self.active_node.parent:
      self.active_node = self.active_node.parent
    else:
      raise ValueError(
          'You have tried to pop out of the highest level in this `SymbolTree`.'
      )

  def drop_scope_down(self, comp_id: int) -> None:
    """Constructs a new scope level for `self`.

    Scope levels in `SymbolTree` correspond to scope-introducing nodes in ASTs;
    that is, either `building_blocks.Block` or `building_blocks.Lambda` nodes.
    Inside of these levels, variables are bound in sequence. The implementer of
    a transformation function needing to interact with scope should never need
    to explicitly walk the scope levels `drop_scope_down` constructs;
    `drop_scope_down` is simply provided for ease of exposing to a traversal
    function.

    Args:
      comp_id: Integer representing a unique key for the
        `building_blocks.ComputationBuildingBlock` which is defines this scope.
        Used to differentiate between scopes which both branch from the same
        point in the tree.
    """
    self.active_node = typing.cast(SequentialBindingNode, self.active_node)
    if self.active_node.children.get(comp_id) is None:
      node = SequentialBindingNode(_BeginScopePointer())
      self._add_child(comp_id, node)
      self._move_to_child(comp_id)
    else:
      self._move_to_child(comp_id)

  def walk_down_one_variable_binding(self):
    """Moves `active_node` to the younger sibling of the current active node.

    This action represents walking from one variable binding in the
    `SymbolTree` to the next, sequentially.

    If there is no such variable binding, then the lower bound variables must
    be accessed via `drop_scope_down`.

    Raises:
      Raises ValueError if there is no such available variable binding.
    """
    self.active_node = typing.cast(SequentialBindingNode, self.active_node)
    if self.active_node.younger_sibling:
      self.active_node = self.active_node.younger_sibling
    else:
      raise ValueError(
          'You have tried to move to a nonexistent variable binding in {}'
          .format(self)
      )

  def ingest_variable_binding(
      self,
      name: Optional[str],
      value: Optional[building_blocks.ComputationBuildingBlock],
  ) -> None:
    """Constructs or updates node in symbol tree as AST is walked.

    Passes `name` and `value` onto the symbol tree's node constructor, with
    `mode` determining how the node being constructed or updated
    relates to the symbol tree's `active_node`.

    If there is no preexisting node in the symbol tree bearing the
    requested relationship to the active node, a new one will be constructed and
    initialized. If there is an existing node, `ingest_variable_binding` checks
    that this node has the correct `payload.name`, and overwrites its
    `payload.value` with the `value` argument.

    Args:
      name: The string name of the `CompTracker` instance we are constructing or
        updating.
      value: Instance of `building_blocks.ComputationBuildingBlock` or `None`,
        as in the `value` to pass to symbol tree's node payload constructor.

    Raises:
      ValueError: If we are passed a name-mode pair such that a
        preexisting node in the symbol tree bears this relationship with
        the active node, but has a different name. This is an indication
        that either a transformation has failed to happen in the symbol tree
        or that we have a symbol tree instance that does not match the
        computation we are currently processing.
    """
    if (name is None or not name) and value is None:
      return

    node = SequentialBindingNode(self.payload_type(name=name, value=value))
    self.active_node = typing.cast(SequentialBindingNode, self.active_node)
    if self.active_node.younger_sibling is None:
      self._add_younger_sibling(node)
      self.walk_down_one_variable_binding()
    else:
      if self.active_node.younger_sibling.payload.name != name:
        raise ValueError(
            'You have a mismatch between your symbol tree and the '
            'computation you are trying to process; your symbol tree is {} '
            'and you are looking for a BoundVariableTracker with name {} '
            'and value {}'.format(self, name, value)
        )
      self.walk_down_one_variable_binding()
      self.active_node.payload.value = value

  def _add_younger_sibling(self, comp_tracker: 'SequentialBindingNode') -> None:
    """Appends comp as younger sibling of current `active_node`."""
    if self._node_ids.get(id(comp_tracker)):
      raise ValueError(
          'Each instance of {} can only appear once in a given symbol tree.'
          .format(self.payload_type)
      )
    self.active_node = typing.cast(SequentialBindingNode, self.active_node)
    if self.active_node.younger_sibling is not None:
      raise ValueError('Ambiguity in adding a younger sibling')
    comp_tracker.set_older_sibling(self.active_node)
    self.active_node.set_younger_sibling(comp_tracker)
    self._node_ids[id(comp_tracker)] = 1

  def _add_child(
      self,
      constructing_comp_id: int,
      comp_tracker: 'SequentialBindingNode',
  ) -> None:
    """Writes `comp_tracker` to children of active node.

    Each `SequentialBindingNode` keeps a `dict` of its children; `_add_child`
    updates the value of this `dict` with key `constructing_comp_id` to be
    `comp_tracker`.

    Notice that `constructing_comp_id` is simply a way of addressing the
    children in this dict; it is not necessarily globally unique, as long
    as it is sufficient to address child scopes.

    Args:
      constructing_comp_id: Key to identify child being constructed from the
        parent scope.
      comp_tracker: Instance of `SequentialBindingNode`, the node to add as a
        child of `active_node`.
    """
    if self._node_ids.get(id(comp_tracker)):
      raise ValueError(
          'Each node can only appear once in a given'
          'symbol tree. You have tried to add {} '
          'twice.'.format(comp_tracker.payload)
      )
    self.active_node = typing.cast(SequentialBindingNode, self.active_node)
    comp_tracker.set_parent(self.active_node)
    self.active_node.add_child(constructing_comp_id, comp_tracker)
    self._node_ids[id(comp_tracker)] = 1

  def _move_to_child(self, comp_id):
    """Moves `active_node` to child of current active node with key `comp_id`.

    Args:
      comp_id: Integer representing the position of the child we wish to update
        `active_node` to point to in a preorder traversal of the AST.

    Raises:
      ValueError: If the active node has no child with the correct id.
    """
    self.active_node = typing.cast(SequentialBindingNode, self.active_node)
    if self.active_node.children.get(comp_id) is not None:
      self.active_node = self.active_node.get_child(comp_id)
    else:
      raise ValueError('You have tried to move to a nonexistent child.')

  def _equal_under_node(self, self_node, other_node):
    """Recursive helper function to check equality of `SymbolTree`s."""
    if self_node is None and other_node is None:
      return True
    if self_node is None or other_node is None:
      return False
    if self_node.payload != other_node.payload:
      return False
    if len(self_node.children) != len(other_node.children):
      return False
    for (_, val_1), (_, val_2) in zip(
        self_node.children.items(), other_node.children.items()
    ):
      # keys not compared to avoid coupling walking logic to `SymbolTree`.
      if not self._equal_under_node(val_1, val_2):
        return False
    return self._equal_under_node(
        self_node.younger_sibling, other_node.younger_sibling
    )

  def __eq__(self, other):
    """Walks to root of `self` and `other` before testing equality of subtrees.

    Args:
      other: Instance of `SymbolTree` to test for equality with `self`.

    Returns:
      Returns `True` if and only if `self` and `other` are the same
      structurally (each node has the same number of children and siblings) and
      each node of `self` compares as equal with the node in the corresponding
      position of `other`.
    """
    if self is other:
      return True
    if not isinstance(other, SymbolTree):
      return NotImplemented
    self_root = _walk_to_root(self.active_node)
    other_root = _walk_to_root(other.active_node)
    return self._equal_under_node(self_root, other_root)

  def __ne__(self, other):
    return not self == other

  def _string_under_node(self, node: 'SequentialBindingNode') -> str:
    """Rescursive helper function to generate string reps of `SymbolTree`s."""
    if node is self.active_node:
      active_node_indicator = '*'
    else:
      active_node_indicator = ''
    symbol_tree_string = '[' + str(node.payload) + active_node_indicator + ']'
    if node.children:
      symbol_tree_string += '->{'
      for _, child_node in node.children.items():
        if not child_node.older_sibling:
          symbol_tree_string += '('
          symbol_tree_string += self._string_under_node(child_node)
          symbol_tree_string += '),('
      symbol_tree_string = symbol_tree_string[:-2]
      symbol_tree_string += '}'
    if node.younger_sibling:
      symbol_tree_string += '-' + self._string_under_node(node.younger_sibling)
    return symbol_tree_string

  def __str__(self):
    """Generates a string representation of this `SymbolTree`.

    First we walk up to the root node, then we walk down
    the tree generating string rep of this symbol tree.

    Returns:
      Returns a string representation of the current `SymbolTree`, with
      the node labeled the active node identified with a *.
    """
    node = self.active_node
    root_node = _walk_to_root(node)
    return self._string_under_node(root_node)


def _walk_to_root(node):
  while node.parent is not None or node.older_sibling is not None:
    while node.older_sibling is not None:
      node = node.older_sibling
    while node.parent is not None:
      node = node.parent
  return node


class SequentialBindingNode:
  """Represents a node in a context tree with sequential-binding semantics.

  `SequentialBindingNode`s are designed to be constructed and pushed into
  a context tree as an AST representing a given computation is walked.

  Each `SequentialBindingNode` holds as payload a variable binding in the AST.
  The node-node relationships encoded by the `SequentialBindingNode` data
  structure determine how the context tree must be walked in order to resolve
  variables and track their values in the AST.

  Parent-child relationships represent relationships between levels of the AST,
  meaning, moving through an AST node which defines a variable scope in preorder
  corresponds to moving from a `SequentialBindingNode` to one of its children,
  and moving through such a node postorder corresponds to moving from a
  `SequentialBindingNode` to its parent.

  Sibling-sibling relationships are particular to sequential binding of
  variables in `building_blocks.Block` constructs; binding
  a new variable in such a construct corresponds to moving from a
  `SequentialBindingNode` to its (unique) younger sibling.
  """

  def __init__(self, payload: BoundVariableTracker):
    """Initializes `SequentialBindingNode`.

    Args:
      payload: Instance of BoundVariableTracker representing the payload of this
        node.
    """
    self.payload = payload
    self._children = collections.OrderedDict()
    self._parent = None
    self._older_sibling = None
    self._younger_sibling = None

  @property
  def parent(self):
    return self._parent

  @property
  def children(self):
    return self._children

  @property
  def older_sibling(self):
    return self._older_sibling

  @property
  def younger_sibling(self):
    return self._younger_sibling

  def set_parent(self, node: 'SequentialBindingNode') -> None:
    """Sets the _parent scope of `self` to the binding embodied by `node`.

    This method should not be assumed to be efficient.

    Args:
      node: Instance of `SequentialBindingNode` to set as parent of `self`.
    """
    self._parent = node

  def set_older_sibling(self, node: 'SequentialBindingNode') -> None:
    """Sets the older sibling scope of `self` to `node`.

    This method should not be assumed to be efficient.

    Args:
      node: Instance of `SequentialBindingNode` to set as older sibling of
        `self`.
    """
    self._older_sibling = node

  def set_younger_sibling(self, node: 'SequentialBindingNode') -> None:
    """Sets the younger sibling scope of `self` to `node`.

    This corresponds to binding a new variable in a
    `building_blocks.Block` construct.

    This method should not be assumed to be efficient.

    Args:
      node: Instance of `SequentialBindingNode` representing this new binding.
    """
    self._younger_sibling = node

  def add_child(self, comp_id: int, node: 'SequentialBindingNode') -> None:
    """Sets the child scope of `self` indexed by `comp_id` to `node`.

    This corresponds to encountering a node in an AST which defines a variable
    scope.

    If a child with this `comp_id` already exists, it is replaced, as in a
    `dict`.

    Args:
      comp_id: The identifier of the computation generating this scope.
      node: Instance of `SequentialBindingNode` representing this new binding.
    """
    self._children[comp_id] = node

  def get_child(self, comp_id):
    """Returns the child of `self` identified by `comp_id` if one exists.

    Args:
      comp_id: Integer used to address child of `self` by position of
        corresponding AST node in a preorder traversal of the AST.

    Returns:
      Instance of `SequentialBindingNode` if an appropriate child of `self`
      exists, or `None`.
    """
    return self._children.get(comp_id)


class ReferenceCounter(BoundVariableTracker):
  """Data container to track number References to a variable in an AST.

  Attributes:
    name: The string name representing the variable whose binding is represented
      by an instance of `ReferenceCounter`.
    value: The value bound to `name`. Can be an instance of
      `building_blocks.ComputationBuildingBlock` or None if this binding is
      simply a placeholder, e.g. in a Lambda.
    count: An integer tracking how many times the variable an instance of
      `ReferenceCounter` represents is referenced in an AST.
  """

  def __init__(self, name, value):
    super().__init__(name, value)
    self.count = 0

  def update(self, value=None):
    del value  # Unused.
    self.count += 1

  def __str__(self):
    return 'Instance count: {}; value: {}; name: {}.'.format(
        self.count, self.value, self.name
    )

  def __repr__(self):
    return str(self)

  def __eq__(self, other):
    if self is other:
      return True
    if not isinstance(other, ReferenceCounter):
      return NotImplemented
    if not super().__eq__(other):
      return False
    return self.count == other.count


def get_unique_names(
    comp: building_blocks.ComputationBuildingBlock,
) -> set[str]:
  """Returns the unique names bound or referred to in `comp`."""
  names = set()

  def _update(comp):
    if isinstance(comp, building_blocks.Block):
      names.update([name for name, _ in comp.locals])
    elif isinstance(comp, building_blocks.Lambda):
      if comp.parameter_type is not None:
        names.add(comp.parameter_name)
    elif isinstance(comp, building_blocks.Reference):
      names.add(comp.name)
    return comp, False

  transform_postorder(comp, _update)
  return names


def get_map_of_unbound_references(
    comp: building_blocks.ComputationBuildingBlock,
) -> dict[building_blocks.ComputationBuildingBlock, set[str]]:
  """Gets a Python `dict` of unbound references in `comp`, keyed by Python `id`.

  Computations that are equal will have the same collections of unbounded
  references, so it is safe to use `comp` as the key for this `dict` even though
  a given computation may appear in many positions in the AST.

  Args:
    comp: The computation building block to parse.

  Returns:
    A Python `dict` of elements where keys are the computations in `comp` and
    values are a Python `set` of the names of the unbound references in the
    subtree of that computation.
  """
  references = {}

  def _update(comp):
    """Updates the Python dict of references."""
    if isinstance(comp, building_blocks.Reference):
      references[comp] = set((comp.name,))
    elif isinstance(comp, building_blocks.Block):
      references[comp] = set()
      names = []
      for name, variable in comp.locals:
        elements = references[variable]
        references[comp].update([e for e in elements if e not in names])
        names.append(name)
      elements = references[comp.result]
      references[comp].update([e for e in elements if e not in names])
    elif isinstance(comp, building_blocks.Call):
      elements = references[comp.function].copy()
      if comp.argument is not None:
        elements.update(references[comp.argument])
      references[comp] = elements
    elif isinstance(comp, building_blocks.Lambda):
      elements = references[comp.result]
      references[comp] = set([e for e in elements if e != comp.parameter_name])
    elif isinstance(comp, building_blocks.Selection):
      references[comp] = references[comp.source]
    elif isinstance(comp, building_blocks.Struct):
      elements = [references[e] for e in comp]
      references[comp] = set(itertools.chain.from_iterable(elements))
    else:
      references[comp] = set()
    return comp, False

  transform_postorder(comp, _update)
  return references



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: federated_language/proto/computation.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'federated_language/proto/computation.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from federated_language.proto import array_pb2 as federated__language_dot_proto_dot_array__pb2
from federated_language.proto import data_type_pb2 as federated__language_dot_proto_dot_data__type__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*federated_language/proto/computation.proto\x12\x12\x66\x65\x64\x65rated_language\x1a\x19google/protobuf/any.proto\x1a$federated_language/proto/array.proto\x1a(federated_language/proto/data_type.proto\"\x84\x05\n\x0b\x43omputation\x12&\n\x04type\x18\x01 \x01(\x0b\x32\x18.federated_language.Type\x12\x34\n\ntensorflow\x18\x02 \x01(\x0b\x32\x1e.federated_language.TensorFlowH\x00\x12\x32\n\tintrinsic\x18\x03 \x01(\x0b\x32\x1d.federated_language.IntrinsicH\x00\x12(\n\x04\x64\x61ta\x18\n \x01(\x0b\x32\x18.federated_language.DataH\x00\x12,\n\x06lambda\x18\x04 \x01(\x0b\x32\x1a.federated_language.LambdaH\x00\x12*\n\x05\x62lock\x18\x05 \x01(\x0b\x32\x19.federated_language.BlockH\x00\x12\x32\n\treference\x18\x06 \x01(\x0b\x32\x1d.federated_language.ReferenceH\x00\x12(\n\x04\x63\x61ll\x18\x07 \x01(\x0b\x32\x18.federated_language.CallH\x00\x12,\n\x06struct\x18\x08 \x01(\x0b\x32\x1a.federated_language.StructH\x00\x12\x32\n\tselection\x18\t \x01(\x0b\x32\x1d.federated_language.SelectionH\x00\x12\x32\n\tplacement\x18\x0b \x01(\x0b\x32\x1d.federated_language.PlacementH\x00\x12.\n\x07literal\x18\x0e \x01(\x0b\x32\x1b.federated_language.LiteralH\x00\x12&\n\x03xla\x18\x0c \x01(\x0b\x32\x17.federated_language.XlaH\x00\x42\r\n\x0b\x63omputationJ\x04\x08\r\x10\x0e\"\x84\x03\n\x04Type\x12\x34\n\x08\x66unction\x18\x01 \x01(\x0b\x32 .federated_language.FunctionTypeH\x00\x12\x30\n\x06struct\x18\x02 \x01(\x0b\x32\x1e.federated_language.StructTypeH\x00\x12\x34\n\x08sequence\x18\x03 \x01(\x0b\x32 .federated_language.SequenceTypeH\x00\x12\x30\n\x06tensor\x18\x04 \x01(\x0b\x32\x1e.federated_language.TensorTypeH\x00\x12\x34\n\x08\x61\x62stract\x18\x05 \x01(\x0b\x32 .federated_language.AbstractTypeH\x00\x12\x36\n\tplacement\x18\x06 \x01(\x0b\x32!.federated_language.PlacementTypeH\x00\x12\x36\n\tfederated\x18\x07 \x01(\x0b\x32!.federated_language.FederatedTypeH\x00\x42\x06\n\x04type\"e\n\x0c\x46unctionType\x12+\n\tparameter\x18\x01 \x01(\x0b\x32\x18.federated_language.Type\x12(\n\x06result\x18\x02 \x01(\x0b\x32\x18.federated_language.Type\"\x87\x01\n\nStructType\x12\x37\n\x07\x65lement\x18\x01 \x03(\x0b\x32&.federated_language.StructType.Element\x1a@\n\x07\x45lement\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\'\n\x05value\x18\x02 \x01(\x0b\x32\x18.federated_language.Type\"9\n\x0cSequenceType\x12)\n\x07\x65lement\x18\x01 \x01(\x0b\x32\x18.federated_language.Type\"]\n\nTensorType\x12+\n\x05\x64type\x18\x01 \x01(\x0e\x32\x1c.federated_language.DataType\x12\x0c\n\x04\x64ims\x18\x02 \x03(\x03\x12\x14\n\x0cunknown_rank\x18\x03 \x01(\x08\"\x1d\n\x0c\x41\x62stractType\x12\r\n\x05label\x18\x01 \x01(\t\"\x81\x01\n\rPlacementSpec\x12\x33\n\x05label\x18\x01 \x01(\x0b\x32\".federated_language.PlacementLabelH\x00\x12.\n\x05value\x18\x02 \x01(\x0b\x32\x1d.federated_language.PlacementH\x00\x42\x0b\n\tplacement\"\x1f\n\x0ePlacementLabel\x12\r\n\x05label\x18\x01 \x01(\t\"\x18\n\tPlacement\x12\x0b\n\x03uri\x18\x01 \x01(\t\"\x82\x01\n\rFederatedType\x12\x34\n\tplacement\x18\x01 \x01(\x0b\x32!.federated_language.PlacementSpec\x12\x11\n\tall_equal\x18\x02 \x01(\x08\x12(\n\x06member\x18\x03 \x01(\x0b\x32\x18.federated_language.Type\"K\n\rPlacementType\x12:\n\x0einstance_label\x18\x01 \x01(\x0b\x32\".federated_language.PlacementLabel\"\xf8\x05\n\nTensorFlow\x12\'\n\tgraph_def\x18\x01 \x01(\x0b\x32\x14.google.protobuf.Any\x12\x15\n\rinitialize_op\x18\x04 \x01(\t\x12!\n\x19session_token_tensor_name\x18\x06 \x01(\t\x12\x39\n\tparameter\x18\x02 \x01(\x0b\x32&.federated_language.TensorFlow.Binding\x12\x36\n\x06result\x18\x03 \x01(\x0b\x32&.federated_language.TensorFlow.Binding\x12:\n\tcache_key\x18\x05 \x01(\x0b\x32\'.federated_language.TensorFlow.CacheKey\x1a\xd8\x01\n\x07\x42inding\x12>\n\x06struct\x18\x01 \x01(\x0b\x32,.federated_language.TensorFlow.StructBindingH\x00\x12>\n\x06tensor\x18\x02 \x01(\x0b\x32,.federated_language.TensorFlow.TensorBindingH\x00\x12\x42\n\x08sequence\x18\x03 \x01(\x0b\x32..federated_language.TensorFlow.SequenceBindingH\x00\x42\t\n\x07\x62inding\x1aH\n\rStructBinding\x12\x37\n\x07\x65lement\x18\x01 \x03(\x0b\x32&.federated_language.TensorFlow.Binding\x1a\x31\n\rTensorBinding\x12\x15\n\x0btensor_name\x18\x01 \x01(\tH\x00\x42\t\n\x07\x62inding\x1a\x62\n\x0fSequenceBinding\x12\x1d\n\x13variant_tensor_name\x18\x02 \x01(\tH\x00\x12\x1f\n\x15graph_def_tensor_name\x18\x03 \x01(\tH\x00\x42\t\n\x07\x62indingJ\x04\x08\x01\x10\x02\x1a\x16\n\x08\x43\x61\x63heKey\x12\n\n\x02id\x18\x01 \x01(\x04J\x04\x08\x07\x10\x08\"\x18\n\tIntrinsic\x12\x0b\n\x03uri\x18\x01 \x01(\t\"Q\n\x06Lambda\x12\x16\n\x0eparameter_name\x18\x01 \x01(\t\x12/\n\x06result\x18\x02 \x01(\x0b\x32\x1f.federated_language.Computation\"\xaf\x01\n\x05\x42lock\x12.\n\x05local\x18\x01 \x03(\x0b\x32\x1f.federated_language.Block.Local\x12/\n\x06result\x18\x02 \x01(\x0b\x32\x1f.federated_language.Computation\x1a\x45\n\x05Local\x12\x0c\n\x04name\x18\x01 \x01(\t\x12.\n\x05value\x18\x02 \x01(\x0b\x32\x1f.federated_language.Computation\"\x19\n\tReference\x12\x0c\n\x04name\x18\x01 \x01(\t\"l\n\x04\x43\x61ll\x12\x31\n\x08\x66unction\x18\x01 \x01(\x0b\x32\x1f.federated_language.Computation\x12\x31\n\x08\x61rgument\x18\x02 \x01(\x0b\x32\x1f.federated_language.Computation\"\x86\x01\n\x06Struct\x12\x33\n\x07\x65lement\x18\x01 \x03(\x0b\x32\".federated_language.Struct.Element\x1aG\n\x07\x45lement\x12\x0c\n\x04name\x18\x01 \x01(\t\x12.\n\x05value\x18\x02 \x01(\x0b\x32\x1f.federated_language.Computation\"K\n\tSelection\x12/\n\x06source\x18\x01 \x01(\x0b\x32\x1f.federated_language.Computation\x12\r\n\x05index\x18\x03 \x01(\x05\"P\n\x04\x44\x61ta\x12\x11\n\x03uri\x18\x01 \x01(\tB\x02\x18\x01H\x00\x12\'\n\x07\x63ontent\x18\x03 \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x42\x06\n\x04\x64\x61taJ\x04\x08\x02\x10\x03\"\x8d\x03\n\x03Xla\x12(\n\nhlo_module\x18\x01 \x01(\x0b\x32\x14.google.protobuf.Any\x12\x32\n\tparameter\x18\x02 \x01(\x0b\x32\x1f.federated_language.Xla.Binding\x12/\n\x06result\x18\x03 \x01(\x0b\x32\x1f.federated_language.Xla.Binding\x1a\x86\x01\n\x07\x42inding\x12\x37\n\x06struct\x18\x01 \x01(\x0b\x32%.federated_language.Xla.StructBindingH\x00\x12\x37\n\x06tensor\x18\x02 \x01(\x0b\x32%.federated_language.Xla.TensorBindingH\x00\x42\t\n\x07\x62inding\x1a\x41\n\rStructBinding\x12\x30\n\x07\x65lement\x18\x01 \x03(\x0b\x32\x1f.federated_language.Xla.Binding\x1a+\n\rTensorBinding\x12\x0f\n\x05index\x18\x01 \x01(\x05H\x00\x42\t\n\x07\x62inding\"3\n\x07Literal\x12(\n\x05value\x18\x01 \x01(\x0b\x32\x19.federated_language.Arrayb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'federated_language.proto.computation_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_DATA'].fields_by_name['uri']._loaded_options = None
  _globals['_DATA'].fields_by_name['uri']._serialized_options = b'\030\001'
  _globals['_COMPUTATION']._serialized_start=174
  _globals['_COMPUTATION']._serialized_end=818
  _globals['_TYPE']._serialized_start=821
  _globals['_TYPE']._serialized_end=1209
  _globals['_FUNCTIONTYPE']._serialized_start=1211
  _globals['_FUNCTIONTYPE']._serialized_end=1312
  _globals['_STRUCTTYPE']._serialized_start=1315
  _globals['_STRUCTTYPE']._serialized_end=1450
  _globals['_STRUCTTYPE_ELEMENT']._serialized_start=1386
  _globals['_STRUCTTYPE_ELEMENT']._serialized_end=1450
  _globals['_SEQUENCETYPE']._serialized_start=1452
  _globals['_SEQUENCETYPE']._serialized_end=1509
  _globals['_TENSORTYPE']._serialized_start=1511
  _globals['_TENSORTYPE']._serialized_end=1604
  _globals['_ABSTRACTTYPE']._serialized_start=1606
  _globals['_ABSTRACTTYPE']._serialized_end=1635
  _globals['_PLACEMENTSPEC']._serialized_start=1638
  _globals['_PLACEMENTSPEC']._serialized_end=1767
  _globals['_PLACEMENTLABEL']._serialized_start=1769
  _globals['_PLACEMENTLABEL']._serialized_end=1800
  _globals['_PLACEMENT']._serialized_start=1802
  _globals['_PLACEMENT']._serialized_end=1826
  _globals['_FEDERATEDTYPE']._serialized_start=1829
  _globals['_FEDERATEDTYPE']._serialized_end=1959
  _globals['_PLACEMENTTYPE']._serialized_start=1961
  _globals['_PLACEMENTTYPE']._serialized_end=2036
  _globals['_TENSORFLOW']._serialized_start=2039
  _globals['_TENSORFLOW']._serialized_end=2799
  _globals['_TENSORFLOW_BINDING']._serialized_start=2328
  _globals['_TENSORFLOW_BINDING']._serialized_end=2544
  _globals['_TENSORFLOW_STRUCTBINDING']._serialized_start=2546
  _globals['_TENSORFLOW_STRUCTBINDING']._serialized_end=2618
  _globals['_TENSORFLOW_TENSORBINDING']._serialized_start=2620
  _globals['_TENSORFLOW_TENSORBINDING']._serialized_end=2669
  _globals['_TENSORFLOW_SEQUENCEBINDING']._serialized_start=2671
  _globals['_TENSORFLOW_SEQUENCEBINDING']._serialized_end=2769
  _globals['_TENSORFLOW_CACHEKEY']._serialized_start=2771
  _globals['_TENSORFLOW_CACHEKEY']._serialized_end=2793
  _globals['_INTRINSIC']._serialized_start=2801
  _globals['_INTRINSIC']._serialized_end=2825
  _globals['_LAMBDA']._serialized_start=2827
  _globals['_LAMBDA']._serialized_end=2908
  _globals['_BLOCK']._serialized_start=2911
  _globals['_BLOCK']._serialized_end=3086
  _globals['_BLOCK_LOCAL']._serialized_start=3017
  _globals['_BLOCK_LOCAL']._serialized_end=3086
  _globals['_REFERENCE']._serialized_start=3088
  _globals['_REFERENCE']._serialized_end=3113
  _globals['_CALL']._serialized_start=3115
  _globals['_CALL']._serialized_end=3223
  _globals['_STRUCT']._serialized_start=3226
  _globals['_STRUCT']._serialized_end=3360
  _globals['_STRUCT_ELEMENT']._serialized_start=3289
  _globals['_STRUCT_ELEMENT']._serialized_end=3360
  _globals['_SELECTION']._serialized_start=3362
  _globals['_SELECTION']._serialized_end=3437
  _globals['_DATA']._serialized_start=3439
  _globals['_DATA']._serialized_end=3519
  _globals['_XLA']._serialized_start=3522
  _globals['_XLA']._serialized_end=3919
  _globals['_XLA_BINDING']._serialized_start=3673
  _globals['_XLA_BINDING']._serialized_end=3807
  _globals['_XLA_STRUCTBINDING']._serialized_start=3809
  _globals['_XLA_STRUCTBINDING']._serialized_end=3874
  _globals['_XLA_TENSORBINDING']._serialized_start=3876
  _globals['_XLA_TENSORBINDING']._serialized_end=3919
  _globals['_LITERAL']._serialized_start=3921
  _globals['_LITERAL']._serialized_end=3972
# @@protoc_insertion_point(module_scope)
